from itertools import islice
from textwrap import dedent

import pytest
from pynmrstar import Entry, Saveframe

from nef_pipelines.lib.sequence_lib import (
    BadResidue,
    chain_code_iter,
    count_residues,
    frame_to_chains,
    get_chain_starts,
    offset_chain_residues,
    sequence_3let_to_sequence_residues,
    sequence_from_frame,
    translate_1_to_3,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.test_lib import path_in_test_data

ABC_SEQUENCE_1LET = "acdefghiklmnpqrstvwy"
ABC_SEQUENCE_3LET = (
    "ALA",
    "CYS",
    "ASP",
    "GLU",
    "PHE",
    "GLY",
    "HIS",
    "ILE",
    "LYS",
    "LEU",
    "MET",
    "ASN",
    "PRO",
    "GLN",
    "ARG",
    "SER",
    "THR",
    "VAL",
    "TRP",
    "TYR",
)

ABC_SEQUENCE_RESIDUES = [
    SequenceResidue("A", i, residue)
    for (i, residue) in enumerate(ABC_SEQUENCE_3LET, start=1)
]


def test_1let_3let():

    assert len(ABC_SEQUENCE_1LET) == 20
    assert len(ABC_SEQUENCE_3LET) == 20

    assert list(ABC_SEQUENCE_3LET) == translate_1_to_3(ABC_SEQUENCE_1LET)


def test_bad_1let_3let():
    BAD_SEQUENCE = "acdefghiklmonpqrstvwy"

    msgs = """\
              unknown residue o
              at residue 12
              sequence:
              acdefghiklmonpqrstvwy
              ^
              """

    msgs = dedent(msgs)
    msgs = msgs.split("\n")

    with pytest.raises(BadResidue) as exc_info:
        translate_1_to_3(BAD_SEQUENCE)

    for msg in msgs:
        assert msg in str(exc_info.value)


def test_3let_sequence_residue():

    sequence_residues = sequence_3let_to_sequence_residues(ABC_SEQUENCE_3LET)

    assert sequence_residues == ABC_SEQUENCE_RESIDUES


def test_3let_sequence_residue_diff_chain():
    sequence_residues = sequence_3let_to_sequence_residues(
        ABC_SEQUENCE_3LET, chain_code="B"
    )

    expected = [
        SequenceResidue("B", residue.sequence_code, residue.residue_name)
        for residue in ABC_SEQUENCE_RESIDUES
    ]

    assert sequence_residues == expected


TEST_DATA_MULTI_CHAIN = """
    save_nef_molecular_system
       _nef_molecular_system.sf_category   nef_molecular_system
       _nef_molecular_system.sf_framecode  nef_molecular_system

       loop_
          _nef_sequence.index
          _nef_sequence.chain_code
          _nef_sequence.sequence_code
          _nef_sequence.residue_name
          _nef_sequence.linking
          _nef_sequence.residue_variant
          _nef_sequence.cis_peptide
          _nef_sequence.ccpn_comment
          _nef_sequence.ccpn_chain_role
          _nef_sequence.ccpn_compound_name
          _nef_sequence.ccpn_chain_comment

         1    A   3    HIS   .   .   .   .   .   Sec5   .
         2    A   4    MET   .   .   .   .   .   Sec5   .
         3    B   5    ARG   .   .   .   .   .   Sec5   .
         4    B   6    GLN   .   .   .   .   .   Sec5   .
         5    C   7    PRO   .   .   .   .   .   Sec5   .

       stop_

    save_

    """


def test_list_chains():

    test_frame = Saveframe.from_string(TEST_DATA_MULTI_CHAIN)
    chains = frame_to_chains(test_frame)

    assert chains == list(["A", "B", "C"])


def test_list_chains_no_chains():
    TEST_DATA = """
    save_nef_molecular_system
       _nef_molecular_system.sf_category   nef_molecular_system
       _nef_molecular_system.sf_framecode  nef_molecular_system

       loop_
          _nef_sequence.index
          _nef_sequence.chain_code
          _nef_sequence.sequence_code
          _nef_sequence.residue_name
          _nef_sequence.linking
          _nef_sequence.residue_variant
          _nef_sequence.cis_peptide
          _nef_sequence.ccpn_comment
          _nef_sequence.ccpn_chain_role
          _nef_sequence.ccpn_compound_name
          _nef_sequence.ccpn_chain_comment

         1    .   3    HIS   .   .   .   .   .   Sec5   .

       stop_

    save_

    """
    test_frame = Saveframe.from_string(TEST_DATA)
    chains = frame_to_chains(test_frame)

    assert chains == list([])


CHAIN_STARTS_TEST_DATA = [
    SequenceResidue(chain_code="A", sequence_code=-1, residue_name="ALA"),
    SequenceResidue(chain_code="A", sequence_code=2, residue_name="ALA"),
    SequenceResidue(chain_code="A", sequence_code="z", residue_name="ALA"),
    SequenceResidue(chain_code="B", sequence_code=2, residue_name="ALA"),
    SequenceResidue(chain_code="B", sequence_code=2, residue_name="ALA"),
    SequenceResidue(chain_code="B", sequence_code=3, residue_name="ALA"),
    SequenceResidue(chain_code="c", sequence_code="y", residue_name="ALA"),
]


def test_count_chains():
    test_frame = Saveframe.from_string(TEST_DATA_MULTI_CHAIN)

    result = {}
    for chain in "ABC":
        result[chain] = count_residues(test_frame, chain)

    EXPECTED = {"A": {"HIS": 1, "MET": 1}, "B": {"ARG": 1, "GLN": 1}, "C": {"PRO": 1}}
    assert result == EXPECTED


def test_get_chain_starts():

    EXPECTED = {
        "A": -1,
        "B": 2,
    }

    assert EXPECTED == get_chain_starts(CHAIN_STARTS_TEST_DATA)


def test_offset_chains():

    EXPECTED = [
        SequenceResidue(chain_code="A", sequence_code=1, residue_name="ALA"),
        SequenceResidue(chain_code="A", sequence_code=4, residue_name="ALA"),
        SequenceResidue(chain_code="A", sequence_code="z", residue_name="ALA"),
        SequenceResidue(chain_code="B", sequence_code=0, residue_name="ALA"),
        SequenceResidue(chain_code="B", sequence_code=0, residue_name="ALA"),
        SequenceResidue(chain_code="B", sequence_code=1, residue_name="ALA"),
        SequenceResidue(chain_code="c", sequence_code="y", residue_name="ALA"),
    ]

    offsets = {"A": 2, "B": -2, "C": 1, "D": 3}

    result = offset_chain_residues(CHAIN_STARTS_TEST_DATA, offsets)

    for pair in zip(EXPECTED, result):
        print(pair)

    assert EXPECTED == result


def test_chain_code_iter_basic():
    result = [chain_code for chain_code in islice(chain_code_iter(), 3)]

    EXPECTED = list("ABC")

    assert EXPECTED == result


def test_chain_code_iter_with_user():
    result = [chain_code for chain_code in islice(chain_code_iter("DE"), 3)]

    EXPECTED = list("DEA")

    assert EXPECTED == result


def test_chain_code_iter_with_exclude():
    result = [chain_code for chain_code in islice(chain_code_iter(exclude="BC"), 3)]

    EXPECTED = list("ADE")

    assert EXPECTED == result


def test_sequence_from_frame():
    path = path_in_test_data(__file__, "multi_chain.nef")

    frames = Entry.from_file(path).get_saveframes_by_category("nef_molecular_system")

    sequence = sequence_from_frame(frames[0], chain_codes_to_select=["B", "C"])

    EXPECTED = [
        SequenceResidue(
            chain_code="B",
            sequence_code=5,
            residue_name="ARG",
        ),
        SequenceResidue(chain_code="B", sequence_code=6, residue_name="GLN"),
        SequenceResidue(chain_code="C", sequence_code=7, residue_name="PRO"),
    ]

    assert sequence == EXPECTED


def test_sequence_from_frame_all():
    path = path_in_test_data(__file__, "multi_chain.nef")

    frames = Entry.from_file(path).get_saveframes_by_category("nef_molecular_system")

    sequence = sequence_from_frame(frames[0])

    EXPECTED = [
        SequenceResidue(chain_code="A", sequence_code=3, residue_name="HIS"),
        SequenceResidue(chain_code="A", sequence_code=4, residue_name="MET"),
        SequenceResidue(chain_code="B", sequence_code=5, residue_name="ARG"),
        SequenceResidue(chain_code="B", sequence_code=6, residue_name="GLN"),
        SequenceResidue(chain_code="C", sequence_code=7, residue_name="PRO"),
    ]

    assert sequence == EXPECTED



from pathlib import Path
from typing import List

import typer
from fyeah import f

from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.sequence_lib import (
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.util import STDIN
from nef_pipelines.transcoders.sparky.importers.peaks import (
    _guess_dimensions_if_not_defined_or_throw,
)
from nef_pipelines.transcoders.xeasy import import_app
from nef_pipelines.transcoders.xeasy.xeasy_lib import parse_peaks


# TODO: this needs to be moved to a library
class XeasyPeakListException(Exception):
    pass


class IncompatibleDimensionTypesException(XeasyPeakListException):
    pass


app = typer.Typer()

DEFAULT_NUCLEI_HELP = (
    "nuclei to use for each dimension, if not defined they are guessed from the assignments"
    "or an error is reported"
)


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    frame_name: str = typer.Option(
        "xeasy_{file_name}",
        "-f",
        "--frame-name",
        help="a templated name for the frame {file_name} will be replaced by input filename",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type peaks.txt", metavar="<XEASY-peaks>.peaks"
    ),
    spectrometer_frequency: float = typer.Option(
        600.123456789, help="spectrometer frequency in MHz"
    ),
):
    """convert xeasy peaks file <XEASY-PEAKS>.peaks to NEF"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    sequence = sequence_from_entry_or_exit(entry)

    entry = pipe(
        entry,
        frame_name,
        file_names,
        sequence,
        spectrometer_frequency=spectrometer_frequency,
    )

    print(entry)


def pipe(
    entry,
    frame_name,
    file_names,
    sequence,
    spectrometer_frequency,
):

    xeasy_frames = []

    residue_type_lookup = sequence_to_residue_name_lookup(sequence)
    for file_name in file_names:

        lines = file_name.open().readlines()

        spectrum_type, dimension_info, peaks = parse_peaks(
            lines, file_name, residue_type_lookup
        )

        dimensions = _guess_dimensions_if_not_defined_or_throw(peaks, dimension_info)

        dimensions = [{"axis_code": dimension.axis_code} for dimension in dimensions]

        file_name = Path(file_name).stem  # used in f method...

        frame_name = f(frame_name)

        frame = peaks_to_frame(
            peaks, dimensions, spectrometer_frequency, frame_code=frame_name
        )

        xeasy_frames.append(frame)

    return add_frames_to_entry(entry, xeasy_frames)



from pathlib import Path

import typer

from nef_pipelines.transcoders.fasta.exporters.sequence import (
    sequence as fasta_sequence,
)
from nef_pipelines.transcoders.mars import export_app


@export_app.command()
def sequence(
    chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain_code",
        help=" single chain to export  [default: 'A']",
        metavar="<CHAIN-CODE>",
    ),
    in_file: Path = typer.Option(
        None, "-i", "--in", help="file to read nef data from", metavar="<NEF-FILE>"
    ),
    output_file: Path = typer.Argument(
        None,
        help="file name to output to [default <ENTRY-ID.fasta>] for stdout use -",
        metavar="<FASTA-SEQUENCE-FILE>",
    ),
):
    """- write a mars sequence file [fasta]"""
    fasta_sequence(chain_code, in_file, output_file)



import os
import string
from collections import Counter
from dataclasses import replace
from enum import auto
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from ordered_set import OrderedSet
from pynmrstar import Entry, Loop, Saveframe
from strenum import LowercaseStrEnum

from nef_pipelines.lib.constants import NEF_UNKNOWN
from nef_pipelines.lib.nef_lib import UNUSED, loop_row_namespace_iter

# from nef_pipelines.lib.nef_lib import loop_to_dataframe
from nef_pipelines.lib.structures import AtomLabel, Linking, SequenceResidue
from nef_pipelines.lib.util import (
    chunks,
    exit_error,
    get_display_file_name,
    is_int,
    read_from_file_or_exit,
    strip_characters_right,
)


class MoleculeType(LowercaseStrEnum):
    PROTEIN = auto()
    RNA = auto()
    DNA = auto()
    CARBOH = auto()


NEF_CHAIN_CODE = "chain_code"
ANY_CHAIN = "_"


def chain_code_iter(
    user_chain_codes: List[str] = "", exclude: List[str] = ()
) -> Iterable[str]:
    """
    Yield chain codes in user chain codes and once exhausted yield any remaining letters of the upper case alphabet
    till they run out. A list of chain codes to not use can be provided...

    Args:
        user_chain_codes (str):  string of dot separated chain codes or a list of chain codes
        exclude: chain codes to not include in the iteration

    Returns:
        Iterable[str] single codes
    """

    ascii_uppercase = list(string.ascii_uppercase)

    seen_codes = set()

    for chain_code in user_chain_codes:
        seen_codes.add(chain_code)
        if chain_code not in exclude:
            yield chain_code

    for chain_code in ascii_uppercase:
        if chain_code not in seen_codes and chain_code not in exclude:
            yield chain_code

    raise ValueError("run out of chain names!")


def get_linking(
    target_sequence: List[SequenceResidue],
    no_chain_start=List[str],
    no_chain_end=List[str],
) -> List[SequenceResidue]:
    """
    get the correct linking for residue in a standard sequenced chain

    Args:
        target_sequence (List[SequenceResidue]) : the sequence
        no_chain_start (List[str]): if the chain isn't in the list cap it at the start
        no_end (List[str]): if the chain isn't in the list cap it at the end

    Returns:
         List[SequenceResidue]: correctly linked residues
    """

    by_chain = {}
    for residue in target_sequence:
        by_chain.setdefault(residue.chain_code, []).append((residue, "middle"))

    result = []
    for chain, residues_and_linkages in by_chain.items():

        if chain not in no_chain_start:
            residues_and_linkages[0] = (residues_and_linkages[0][0], "start")

        if chain not in no_chain_end:
            residues_and_linkages[-1] = (residues_and_linkages[-1][0], "end")

        result.extend(residues_and_linkages)

    return result


def sequence_to_nef_frame(
    sequences: List[SequenceResidue],
    no_chain_start: Union[List[str], Tuple[str]] = (),
    no_chain_end: Union[List[str], Tuple[str]] = (),
) -> Saveframe:
    """

    Args:
        sequences (List[SequenceResidue]): the sequence as a list of sequence residues [will be sorted]
        no_chain_start (List[str]): if the chain isn't in the list cap it at the start
        no_chain_end (List[str]): if the chain isn't in the list cap it at the end

    Returns:
        Saveframe: a NEF saveframe
    """

    sequences = sorted(sequences)

    category = "nef_molecular_system"

    frame_code = f"{category}"

    nef_frame = Saveframe.from_scratch(frame_code, category)

    nef_frame.add_tag("sf_category", category)
    nef_frame.add_tag("sf_framecode", frame_code)

    nef_loop = Loop.from_scratch("nef_sequence")
    nef_frame.add_loop(nef_loop)

    tags = (
        "index",
        "chain_code",
        "sequence_code",
        "residue_name",
        "linking",
        "residue_variant",
        "cis_peptide",
    )

    nef_loop.add_tag(tags)

    # TODO need tool to set ionisation correctly
    residue_and_linkages = get_linking(sequences, no_chain_start, no_chain_end)

    for index, (sequence_residue, linking) in enumerate(residue_and_linkages):

        data = [
            {
                "index": index + 1,
                "chain_code": sequence_residue.chain_code,
                "sequence_code": sequence_residue.sequence_code,
                "residue_name": sequence_residue.residue_name.upper(),
                "linking": linking,
                "residue_variant": NEF_UNKNOWN,
                "cis_peptide": NEF_UNKNOWN,
            }
        ]
        nef_loop.add_data(data)

    return nef_frame


class MoleculeTypes(LowercaseStrEnum):
    PROTEIN = auto()
    DNA = auto()
    RNA = auto()
    CARBOH = auto()
    LIGAND = auto()
    OTHER = auto()


TRANSLATIONS_3_1_DNA = {"DG": "G", "DC": "C", "DT": "T", "DA": "A"}
TRANSLATIONS_1_3_DNA = {value: key for (key, value) in TRANSLATIONS_3_1_DNA.items()}

TRANSLATIONS_3_1_RNA = {"G": "G", "C": "C", "A": "A", "U": "U", "I": "I"}
TRANSLATIONS_1_3_RNA = {value: key for (key, value) in TRANSLATIONS_3_1_RNA.items()}

TRANSLATIONS_3_1_PROTEIN = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLU": "E",
    "GLN": "Q",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V",
}
TRANSLATIONS_1_3_PROTEIN = {
    value: key for (key, value) in TRANSLATIONS_3_1_PROTEIN.items()
}

TRANSLATIONS_1_3 = {
    MoleculeTypes.PROTEIN: TRANSLATIONS_1_3_PROTEIN,
    MoleculeTypes.DNA: TRANSLATIONS_1_3_DNA,
    MoleculeTypes.RNA: TRANSLATIONS_1_3_RNA,
    MoleculeTypes.CARBOH: None,
    MoleculeTypes.OTHER: None,
    MoleculeTypes.LIGAND: None,
}
TRANSLATIONS_3_1 = {
    MoleculeTypes.PROTEIN: TRANSLATIONS_3_1_PROTEIN,
    MoleculeTypes.DNA: TRANSLATIONS_3_1_DNA,
    MoleculeTypes.RNA: TRANSLATIONS_3_1_RNA,
    MoleculeTypes.CARBOH: None,
    MoleculeTypes.OTHER: None,
    MoleculeTypes.LIGAND: None,
}


class BadResidue(Exception):
    """
    Bad residue found in a sequence
    """

    def __init__(self, residue_name, error_pos, sequence_string):

        self.residue_name = residue_name
        self.sequence_string = sequence_string
        self.error_pos = error_pos

    def __str__(self):
        try:
            terminal_width = os.get_terminal_size().columns
        except OSError:
            terminal_width = 80

        error_pos_string = [
            " ",
        ] * len(self.sequence_string)
        error_pos_string[self.error_pos] = "^"
        error_pos_string = "".join(error_pos_string)

        msg = f"""\
        unknown residue {self.residue_name}
        at residue {self.error_pos+1}
        sequence:
        """
        msg = dedent(msg)
        sequence_string_chunks = list(chunks(self.sequence_string, terminal_width - 10))
        pos_string_chunks = list(chunks(error_pos_string, terminal_width - 10))
        for sequence_chunk, pos_chunk in zip(sequence_string_chunks, pos_string_chunks):
            msg += f"{''.join(sequence_chunk)}\n"
            msg += f"{''.join(pos_chunk)}\n"

        return msg


def translate_1_to_3(
    sequence: str, molecule_type=MoleculeTypes.PROTEIN, unknown: Optional[str] = None
) -> List[str]:
    """
    Translate a 1 letter sequence to a 3 letter sequence
    Args:
        sequence (str): 1 letter sequence
        molecule_type (MoleculeTypes): type of molecule to translate residue types
        unknown (Optional[str]): optional name for residues if they are unknown, if set no error is raised if a
                                 1 letter residue name is not recognised
    Returns List[str]:
        a list of 3 residue codes
        :param type:

    """

    translations = TRANSLATIONS_1_3[molecule_type]

    result = []
    for i, residue_name_1let in enumerate(sequence):
        if residue_name_1let == " ":
            continue
        original_residue_name_1let = residue_name_1let
        residue_name_1let = residue_name_1let.upper()
        if translations is None:
            result.append(residue_name_1let)
        elif residue_name_1let in translations:
            result.append(translations[residue_name_1let])
        else:
            if unknown:
                result.append(unknown)
            else:
                raise BadResidue(original_residue_name_1let, i, sequence)

    return result


def translate_3_to_1(
    sequence: List[str], molecule_type=MoleculeTypes.PROTEIN
) -> List[str]:
    """

    Translate a 3 letter sequence to a 1 letter sequence
    Args:
        sequence (str): 3 letter sequence
        molecule_type (MoleculeTypes): type of molecule to translate residue types

    Returns List[str]:
        a list of 1 residue codes

    :param sequence:
    :return:
    """

    translations = TRANSLATIONS_3_1[molecule_type]

    result = []
    for i, residue_name in enumerate(sequence, start=1):
        residue_name = residue_name.upper()
        if translations is None:
            if len(residue_name) == 1:
                result.append(residue_name)
            else:
                msg = f"""
                    it isn't possible to translate the residue name {residue_name} to a 1 letter code
                    the molecule type {molecule_type} doesn't have one letter codes and the residue name: {residue_name}
                    is longer than one letter and so can't be passed through
                """
                exit_error(msg)
        if residue_name in translations:
            result.append(translations[residue_name])
        else:
            msg = f"""
            unknown residue {residue_name}
            in sequence {chunks(' '.join(sequence), 10)}
            at residue number {i}
            """
            exit_error(msg)

    return result


def sequence_3let_to_sequence_residues(
    sequence_3let: List[str], chain_code: str = "A"
) -> List[SequenceResidue]:
    """
    Translate a list of 3 residue sequence codes to SequenceResidues
    Args:
        sequence_3let (List([str]): list of 3 letter residue names
        chain_code (str): the chain code [defaults to A]
        offset (int): the sequence offset [defaults to 0]

    Returns List[SequenceResidue]:
        the sequence as a list of SequenceResidues

    """

    return [
        SequenceResidue(chain_code, i, residue)
        for (i, residue) in enumerate(sequence_3let, start=1)
    ]


def sequence_residues_to_sequence_3let(
    sequence: List[SequenceResidue], chain_code: str = "A"
) -> List[str]:
    """
    Translate a list of SequenceResidues to 3 letter sequence codes t
    Args:
        sequence (List([SequenceResidues]): list of residues
        chain_code (str): the chain code [defaults to A] only residues fromthis chain will be converted

    Returns List[str]:
        the sequence as a list of 3 letter residues

    """
    return [
        residue.residue_name for residue in sequence if residue.chain_code == chain_code
    ]


def frame_to_chains(sequence_frame: Saveframe) -> List[str]:
    """
     a nef molecular systems list the chains found

    :param sequence_frames: a nef molecular system save frame
    :return: a list of chain_codes
    """

    chains = set()
    for loop in sequence_frame.loop_dict.values():
        for row in loop_row_namespace_iter(loop):
            chains.add(row.chain_code)

    if NEF_UNKNOWN in chains:
        chains.remove(NEF_UNKNOWN)

    chains = sorted(chains)

    return chains


def count_residues(sequence_frame: Saveframe, chain_code: str) -> Dict[str, int]:
    """
    given a nef molecular system and a chain list the number of residues of each type in the chain

    :param sequence_frames: a nef molecular system save frame
    :return: a list of chain_codes
    """

    residue_number_to_residue_name = {}
    for row in loop_row_namespace_iter(sequence_frame.loops[0]):
        if row.chain_code == chain_code:
            residue_number_to_residue_name[row.sequence_code] = row.residue_name

    result = Counter()
    for residue_name in sorted(residue_number_to_residue_name.values()):
        result[residue_name] += 1

    return result


def get_sequence_or_exit(
    file_name: Path = Path("-"),
) -> Optional[List[SequenceResidue]]:
    """
    read a sequence from a nef molecular system on a file or stdin or exit, return a list of residues

    file_name: a path to read the nef from Path('-') read from stdin
    :return: a list of parsed residues and chains, in the order they were read
    """
    try:
        result = get_sequence(file_name)
    except Exception as e:
        msg = "couldn't read sequence from nef input stream, either no stream or no nef molecular system frame"
        exit_error(msg, e)

    if len(result) == 0:
        msg = "no sequence read from nef molecular system in nef input stream"
        exit_error(msg)

    return result


def get_sequence(file_name: Path = Path("-")) -> List[SequenceResidue]:
    """
    read a sequence from a nef molecular system on stdin and return a list of residues

    file_name: a path to read the nef from Path('-') read from stdin
    can raise Exceptions

    :return:a list of parsed residues and chains, in the order they were read
    """

    text = read_from_file_or_exit(file_name, "sequence from NEF molecular system")

    display_file_name = get_display_file_name(file_name)

    try:
        entry = Entry.from_string(text)
    except Exception as e:
        msg = f"couldn't parse text in {display_file_name} as NEF data"
        exit_error(msg, e)

    return sequence_from_entry(entry)


def sequence_from_entry_or_exit(entry: Entry) -> List[SequenceResidue]:
    sequence = sequence_from_entry(entry)

    if not sequence:
        msg = "did you read a sequence? a sequence is required and was not read"
        exit_error(msg)

    return sequence


def sequence_from_entry(entry) -> List[SequenceResidue]:

    result = []
    if entry is not None:
        frames = entry.get_saveframes_by_category("nef_molecular_system")

        if frames:
            result = sequence_from_frame(frames[0])
    return result


def sequence_from_frame(
    frame: Saveframe, chain_codes_to_select: Union[str, List[str]] = ANY_CHAIN
) -> List[SequenceResidue]:
    """
    read sequences from a nef molecular system save frame
    can raise Exceptions

    :param frame: the save frame to read residues from, must have category nef molecular system
    :param chain_code_to_select: the chain codes to select this can be either a string or list of strings,
                                 any chain is selected using the instance of the constant string ANY_CHAIN [default]
    :return: a list of parsed residues and chains, in the order they were read
    """

    if chain_codes_to_select is not ANY_CHAIN and isinstance(
        chain_codes_to_select, str
    ):
        chain_codes_to_select = set(
            [
                chain_codes_to_select,
            ]
        )
    elif chain_codes_to_select is not ANY_CHAIN:
        chain_codes_to_select = set(chain_codes_to_select)

    residues = OrderedSet()

    if frame.category != "nef_molecular_system":
        raise Exception(
            f"sequences can only be read from nef molecular system frames, the category of the provided frame "
            f"was {frame.category}"
        )

    loop = frame.loops[0]

    chain_code_index = loop.tag_index("chain_code")
    sequence_code_index = loop.tag_index("sequence_code")
    residue_name_index = loop.tag_index("residue_name")
    linking_index = loop.tag_index("linking")
    residue_variant_index = loop.tag_index("residue_variant")

    for line in loop:
        chain_code = line[chain_code_index]
        if chain_codes_to_select is ANY_CHAIN or chain_code in chain_codes_to_select:
            sequence_code = line[sequence_code_index]
            sequence_code = (
                int(sequence_code) if is_int(sequence_code) else sequence_code
            )
            residue_name = line[residue_name_index]
            linking = (
                Linking[line[linking_index].upper()]
                if line[linking_index] != NEF_UNKNOWN
                else None
            )
            residue_variants = line[residue_variant_index].split(",")
            residue_variants = (
                ()
                if residue_variants
                == [
                    UNUSED,
                ]
                else tuple(residue_variants)
            )

            residue = SequenceResidue(
                chain_code=chain_code,
                sequence_code=sequence_code,
                residue_name=residue_name,
                linking=linking,
                variants=residue_variants,
            )
            residues.append(residue)

    return list(residues)


def sequence_3let_to_res(
    sequence_3_let: List[str], chain_code: str, start: int = 1
) -> List[SequenceResidue]:
    """
    convert a list of 3 letter residue names to a list of sequence residues
    :param sequence_3_let: 3 letter names
    :param chain_code: chain code to use
    :param start:  start of the chain [default is 1]
    :return: a list of sequeence residues
    """
    result = set()
    for res_number, residue_name in enumerate(sequence_3_let, start=start):
        result.add(SequenceResidue(chain_code, res_number, residue_name))

    return list(result)


def get_chain_starts(residues: List[SequenceResidue]) -> Dict[str, int]:
    """
    from a list of residues get the lowest residue number for each chain ignoring any sequence codes that
    can't be converted to an integer

    :param residues:  a list of residues from one ore more chains
    :return: a dictionary of chain starts by chain_code
    """

    chain_residue_numbers = {}

    for residue in residues:

        if is_int(residue_number := residue.sequence_code):
            chain_residue_numbers.setdefault(residue.chain_code, []).append(
                residue_number
            )

    return {
        chain_code: min(residue_numbers)
        for chain_code, residue_numbers in chain_residue_numbers.items()
    }


def sequence_to_chains(residues: List[SequenceResidue]) -> List[str]:
    """
    from a list of residues get chain_codes

    :param residues:  a list of residues from one ore more chains
    :return: a list of chain_codes
    """

    chain_codes = set()

    for residue in residues:

        chain_codes.add(residue.chain_code)

    return sorted(tuple(chain_codes))


def offset_chain_residues(
    residues: List[SequenceResidue], chain_residue_offsets: Dict[str, int]
) -> List[SequenceResidue]:
    """
    take a list of residues and offset residue sequence_codes by any offsets in chains if the
    sequence_code can be converted to an integer and the chain_code matches

    :param residues:  a list of residues from one ore more chains
    :return: a list of residues
    """

    result = []
    for residue in residues:
        if residue.chain_code in chain_residue_offsets and is_int(
            residue.sequence_code
        ):
            offset = chain_residue_offsets[residue.chain_code]
            result.append(
                replace(residue, sequence_code=residue.sequence_code + offset)
            )
        else:
            result.append(residue)

    return result


def sequence_3let_list_from_sequence(
    sequence: List[SequenceResidue], count: int = 10, chain_code="A"
) -> List[str]:
    """
    convert a list of Residues to a set of lines containing space separated residue names
    [default 10] per line

    :param sequence: a sequence
    :param count: number of residues names to display per line
    :return: a list of lines as strings
    """

    residue_names = [
        residue.residue_name for residue in sequence if residue.chain_code == chain_code
    ]
    rows = chunks(residue_names, count)

    row_strings = []
    for row in rows:
        row_strings.append(" ".join(row))

    return row_strings


def make_chunked_sequence_1let(
    sequence_1_let: List[str], sub_chunk: int = 10, line_length: int = 100
) -> List[str]:
    """
    convert a list of strings into a chunked list of strings with (by de]fault) every 100 sting being
    separated by a new line [line_length] and sequences of strings being separated by a space
    :param sequence_1_let: a list of strings typically 1 single letter
    :param sub_chunk: how often to separate strings by a space
    :param line_length: how many strings need to be seen before a new line is added
    :return: a list of formatted strings one per line
    """

    rows = chunks(sequence_1_let, line_length)

    row_strings = []
    for row in rows:
        row_chunks = list(chunks(row, sub_chunk))
        row_string = ["".join(chunk) for chunk in row_chunks]
        row_strings.append(" ".join(row_string))

    return row_strings


def replace_chain_in_atom_labels(
    atom_labels: List[AtomLabel], chain_code: str
) -> List[AtomLabel]:
    """
    replace the chain in a list of AtomLabels, not this operation is not inplace new AtomLabels are returned

    :param atom_labels: a list of AtomLabels
    :param chain_code: the new chain_code
    :return: the updated AtomLabels
    """
    result = []

    for selection in atom_labels:
        new_residue = replace(selection.residue, chain_code=chain_code)
        result.append(replace(selection, residue=new_residue))

    return result


def get_residue_name_from_lookup(
    chain_code: str,
    sequence_code: Union[str, int],
    lookup_table: Dict[Union[str, int], str],
) -> str:
    """
    get a residue name from a chain_code and sequence_code, using the provided lookup table. The sequence code is
    treated being as either  a string or an int, strs are checked first
    :param chain_code: the chain code
    :param sequence_code: the residue code treated as both an int and a string
    :param lookup_table: the table of residues names indexed by chain_code_sequence_code
    :return:
    """

    seq_key = chain_code, str(sequence_code)

    residue_name = lookup_table.get(seq_key, None)

    if not residue_name:
        if is_int(sequence_code):
            seq_key = chain_code, int(sequence_code)

    residue_name = lookup_table.get(seq_key, UNUSED)

    if residue_name is not None:
        residue_name = residue_name.upper()

    return residue_name


def sequence_to_residue_name_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:
    """
    build a lookup table from sequence_code, chain_code to residue_type from a list of sequence residues
    :param sequence: a list of sequence residues
    :return: the lookup table
    """

    result: Dict[Tuple[str, int], str] = {}
    if len(sequence) > 0:
        result = {
            (residue.chain_code, str(residue.sequence_code)): residue.residue_name
            for residue in sequence
        }

    return result


def atom_sort_key(item: AtomLabel) -> Tuple[Any, ...]:
    """

    key based sorter for atom labels that enables them to be sorted in a sensible order including sequence codes
    which are not numeric...

    :param item: an atom label
    :return: a tuple baed on the atom label which gives the correct sort order
    """

    try:
        items = []
        residue = item.residue
        chain_code = residue.chain_code.strip()
        prefix, numbers = strip_characters_right(chain_code, string.digits)

        if numbers == "":
            numbers = 0, ""
        if is_int(numbers):
            numbers = int(numbers), ""
        else:
            numbers = 0, numbers

        if UNUSED in prefix:
            order = 2
        elif "@" in prefix or "#" in prefix:
            order = 1
        else:
            order = 0

        items.append([order, prefix, *numbers])

        sequence_code = str(residue.sequence_code).strip()
        prefix, numbers = strip_characters_right(sequence_code, (string.digits + "+-"))

        if is_int(numbers):
            numbers = [int(numbers), "", 0]
        elif "+" in numbers or "-" in numbers:
            for splitter in "+-":
                if splitter in numbers:
                    numbers = numbers.rsplit(splitter, maxsplit=1)
                    numbers = [numbers[0], splitter, numbers[1]]
                    break
            if is_int(numbers[0]) and is_int(numbers[-1]):
                numbers[0] = int(numbers[0])
                numbers[-1] = int(numbers[-1])
        else:
            numbers = [0, numbers, 0]

        if UNUSED in prefix:
            order = 2
        elif "@" in prefix or "#" in prefix:
            order = 1
        else:
            order = 0

        items.append([order, prefix, *numbers])

        items.append(residue.residue_name.lower())
        items.append(item.atom_name.lower())
    except Exception as e:
        msg = f"""
            I can't sort {item} because it doesn't seem to be behaving like an AtomLabel...
            [{e}]
        """
        raise ValueError(msg)

    return items


def exit_if_chain_not_in_sequence(chain_code: str, entry: Entry, file_name: str):

    nef_sequence = sequence_from_entry(entry)

    nef_chain_codes = sequence_to_chains(nef_sequence)

    if chain_code not in nef_chain_codes:
        msg = f"""
            The chain code {chain_code} was not found in the input chain codes in the nef stream
            with entry name {entry.entry_id}  in file {file_name}
            the chain codes found were {','.join(nef_chain_codes)}
        """

        exit_error(msg)



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.frames.delete import delete

runner = CliRunner()
app = typer.Typer()
app.command()(delete)

EXPECTED_DELETE_CATEGORY = """\
        data_pales_test

        save_nef_rdc_restraint_list_test_1
           _nef_rdc_restraint_list.sf_category     nef_rdc_restraint_list
           _nef_rdc_restraint_list.sf_framecode    nef_rdc_restraint_list_test_1
           _nef_rdc_restraint_list.potential_type  log-normal

           loop_
              _nef_rdc_restraint.index
              _nef_rdc_restraint.restraint_id
              _nef_rdc_restraint.restraint_combination_id
              _nef_rdc_restraint.chain_code_1
              _nef_rdc_restraint.sequence_code_1
              _nef_rdc_restraint.residue_name_1
              _nef_rdc_restraint.atom_name_1
              _nef_rdc_restraint.chain_code_2
              _nef_rdc_restraint.sequence_code_2
              _nef_rdc_restraint.residue_name_2
              _nef_rdc_restraint.atom_name_2
              _nef_rdc_restraint.weight
              _nef_rdc_restraint.target_value
              _nef_rdc_restraint.target_value_uncertainty

             1   1   .   A   2   TRP   H   A   21   TRP   N   2   -5.2   0.33
             2   2   .   A   3   GLY   H   A   22   GLY   N   3   3.1    0.4

           stop_

        save_
    """


# noinspection PyUnusedLocal
def test_delete_type(clear_cache):

    path = path_in_test_data(__file__, "pales_test_1.nef")

    input_stream = open(path).read()

    result = run_and_report(app, ["-c", "mol"], input=input_stream)

    assert_lines_match(EXPECTED_DELETE_CATEGORY, result.stdout)


EXPECTED_DELETE_NAME = """\
    data_pales_test

    save_nef_molecular_system
       _nef_molecular_system.sf_category   nef_molecular_system
       _nef_molecular_system.sf_framecode  nef_molecular_system

       loop_
          _nef_sequence.index
          _nef_sequence.chain_code
          _nef_sequence.sequence_code
          _nef_sequence.residue_name
          _nef_sequence.linking
          _nef_sequence.residue_variant
          _nef_sequence.cis_peptide

         1   A   1   ALA   start    .   .
         2   A   2   TRP   middle   .   .
         3   A   3   GLY   middle   .   .

       stop_

    save_

"""


# noinspection PyUnusedLocal
def test_delete_name(clear_cache):

    path = path_in_test_data(__file__, "pales_test_1.nef")

    input_stream = open(path).read()

    result = run_and_report(app, ["test_1"], input=input_stream)

    assert_lines_match(EXPECTED_DELETE_NAME, result.stdout)



import pytest
from pyparsing import ParseException

from nef_pipelines.lib.structures import AtomLabel, SequenceResidue
from nef_pipelines.lib.test_lib import NOQA_E501, path_in_test_data
from nef_pipelines.transcoders.xplor.xplor_lib import (
    XPLOR_COMMENT,
    XPLORParseException,
    _atom_factor,
    _dihedral_restraint,
    _dihedral_restraints,
    _distance_restraints,
    _get_approximate_restraint_strings,
    _get_selection_expressions_from_selection,
    _get_single_atom_selection,
    _NamedToken,
    _residue_factor,
    _segid_factor,
    _selection,
)


def test_parse_residue_factor():
    assert _residue_factor.parseString("residue  20")[0].get_name() == "resid"
    assert _residue_factor.parseString("residue  20")[0][0] == 20
    assert _residue_factor.parseString("residue  -20")[0][0] == -20


def test_parse_atom_factor():
    assert _atom_factor.parseString("name  HD1#")[0].get_name() == "atom"
    assert _atom_factor.parseString("name  HD1#")[0][0] == "HD1#"


def test_parse_segid_factor():
    assert _segid_factor.parseString("segid  HDAA")[0].get_name() == "segid"
    assert _segid_factor.parseString("segid  HDAA")[0][0] == "HDAA"


def test_parse_selection_name():
    TEST_DATA = "(segid AAAA and resid 23 and  (name HA or name HB))"

    selection_result = _selection.parseString(TEST_DATA)

    assert selection_result.getName() == "selection"


def test_simple_selection_no_segid():
    TEST_DATA = "(resid 23 and name HA)"
    result = _selection.parseString(TEST_DATA)

    assert str(result) == "[[[resid : 23, atom : HA]]]"
    assert result.get("selection")[0][0][0][0] == 23
    assert result.get("selection")[0][0][1][0] == "HA"


def test_parse_selection_multiple_names():
    TEST_DATA = "(segid AAAA and resid 23 and  (name HA or name HB))"

    result = _selection.parseString(TEST_DATA)

    assert str(result) == "[[[segid : AAAA, resid : 23, [[atom : HA], [atom : HB]]]]]"

    assert result[0][0][-1][0][0][0] == "HA"
    assert result[0][0][-1][1][0][0] == "HB"


def test_parse_dihedral():
    SEL_1 = "(SEGID AAAA and resid 10 and name HA)"
    SEL_2 = "(SEGID BBBB and resid 11 and name CA)"
    SEL_3 = "(SEGID CCCC and resid 12 and name C)"
    SEL_4 = "(SEGID DDDD and resid 13 and name O)"

    TEST_DATA = f"ASSI {SEL_1} {SEL_2} {SEL_3} {SEL_4}  1.234 4.567 7.8910 2"

    result = _dihedral_restraint.parseString(TEST_DATA)[0]
    assert result.get("atoms_1").as_list() == [
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 10),
                _NamedToken("atom", "HA"),
            ]
        ]
    ]
    assert result.get("atoms_2").as_list() == [
        [
            [
                _NamedToken("segid", "BBBB"),
                _NamedToken("resid", 11),
                _NamedToken("atom", "CA"),
            ]
        ]
    ]
    assert result.get("atoms_3").as_list() == [
        [
            [
                _NamedToken("segid", "CCCC"),
                _NamedToken("resid", 12),
                _NamedToken("atom", "C"),
            ]
        ]
    ]
    assert result.get("atoms_4").as_list() == [
        [
            [
                _NamedToken("segid", "DDDD"),
                _NamedToken("resid", 13),
                _NamedToken("atom", "O"),
            ]
        ]
    ]

    assert result.get("energy_constant") == 1.234
    assert result.get("angle") == 4.567
    assert result.get("range") == 7.8910
    assert result.get("exponent") == 2


def test_read_multiple_dihedral_restraints():

    TEST_DATA = open(path_in_test_data(__file__, "test_2_dihedrals.tbl")).read()

    # print(TEST_DATA)

    result = _dihedral_restraints.ignore(XPLOR_COMMENT).parseString(TEST_DATA)

    assert len(result) == 2

    print(result[0].as_list())
    assert result[0].as_list() == [
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 1),
                _NamedToken("atom", "C"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 2),
                _NamedToken("atom", "N"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 2),
                _NamedToken("atom", "CA"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 2),
                _NamedToken("atom", "C"),
            ]
        ],
        1.0,
        -45.7,
        120.5,
        2,
    ]

    assert result[1].as_list() == [
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 2),
                _NamedToken("atom", "N"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 2),
                _NamedToken("atom", "CA"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 2),
                _NamedToken("atom", "C"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 3),
                _NamedToken("atom", "N"),
            ]
        ],
        1.0,
        65.4,
        120.7,
        2,
    ]


def test_read_3_dihedral_restraints_incomplete():

    TEST_DATA = open(path_in_test_data(__file__, "test_2_dihedrals.tbl")).read()

    TEST_DATA = f"""\
        {TEST_DATA}
        assign (
    """

    with pytest.raises(ParseException) as e_info:
        _dihedral_restraints.ignore(XPLOR_COMMENT).parseString(
            TEST_DATA, parse_all=True
        )

    assert "Expected end of text, found 'assign'" in str(e_info.value)


def test_read_multiple_noe_restraints():

    TEST_DATA = open(path_in_test_data(__file__, "test_2_noes.tbl")).read()

    result = _distance_restraints.ignore(XPLOR_COMMENT).parseString(TEST_DATA)

    assert len(result) == 2

    assert result[0].as_list() == [
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 3),
                _NamedToken("atom", "HN"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 3),
                _NamedToken("atom", "HDA#"),
            ]
        ],
        3.6,
        3.0,
        2.4,
        "",
    ]
    assert result[1].as_list() == [
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 7),
                _NamedToken("atom", "HN"),
            ]
        ],
        [
            [
                _NamedToken("segid", "AAAA"),
                _NamedToken("resid", 7),
                _NamedToken("atom", "HD1#"),
            ]
        ],
        4.0,
        3.3,
        2.7,
        "",
    ]


def test_get_single_atom_selection_correct():
    TEST_DATA = "(segid AAAA and resid 1 and  name HA)"

    selected_atoms = _selection.parseString(TEST_DATA)

    result = _get_single_atom_selection(
        selected_atoms, residue_types={("AAAA", "1"): "Ala"}
    )

    EXPECTED = AtomLabel(
        atom_name="HA",
        residue=SequenceResidue(chain_code="AAAA", sequence_code=1, residue_name="Ala"),
    )

    assert result == EXPECTED


def test_get_single_atom_selection_multiple_selections():

    TEST_DATA = "((resid 1 and  name HA and segid AAAA) or (resid 2 and  name HB and segid BBBB))"

    selected_atoms = _selection.parseString(TEST_DATA)

    with pytest.raises(XPLORParseException):
        _get_single_atom_selection(selected_atoms, residue_types={("AAAA", "1"): "Ala"})


def test_convert_selection_to_selection_expressions_double():
    selection_text = (
        "((segid AAAA and resid 1 and name HA) or (segid BBBB and resid 2 and name HB))"
    )
    result = _selection.parseString(selection_text, parse_all=True)

    selections = _get_selection_expressions_from_selection(result, selection_text)
    assert len(selections) == 2
    assert str(selections[0]) == "[[segid : AAAA, resid : 1, atom : HA]]"
    assert str(selections[1]) == "[[segid : BBBB, resid : 2, atom : HB]]"


def test_convert_selection_to_selection_expressions_single():
    selection_text = "(segid AAAA and resid 1 and name HA)"
    result = _selection.parseString(selection_text, parse_all=True)

    selections = _get_selection_expressions_from_selection(result, selection_text)

    assert len(selections) == 1
    assert str(selections[0]) == "[[segid : AAAA, resid : 1, atom : HA]]"


def test_convert_selection_to_selection_expressions_double_sub_selection():
    selection_text = (
        "((segid AAAA and resid 1 and name HA) or "
        "((segid AAAA or segid BBBB) and (resid 1 or resid 2) and (name HA or name HB)))"
    )

    result = _selection.parseString(selection_text, parse_all=True)

    selections = _get_selection_expressions_from_selection(result, selection_text)

    assert len(selections) == 2
    assert str(selections[0]) == "[[segid : AAAA, resid : 1, atom : HA]]"
    assert (
        str(selections[1]) == "[["
        "[[segid : AAAA], [segid : BBBB]], "
        "[[resid : 1], [resid : 2]], "
        "[[atom : HA], [atom : HB]]"
        "]]"
    )


def test_get_approximate_restraint():
    dihedrals_path = path_in_test_data(__file__, "test_2_dihedrals.tbl")

    dihdedral_restraints = open(dihedrals_path, "r").read()

    result = _get_approximate_restraint_strings(dihdedral_restraints)

    EXPECTED = [
        """# noqa: E501 \
           assign  (segid AAAA and resid    1 and name C    ) (segid AAAA and resid    2 and name N    )
                   (segid AAAA and resid    2 and name CA   ) (segid AAAA and resid    2 and name C    )    1.0 -45.7   120.5 2
        """.replace(
            NOQA_E501, ""
        ),
        """# noqa: E501 \
           assign  (segid AAAA and resid    2 and name N    ) (segid AAAA and resid    2 and name CA   )
                   (segid AAAA and resid    2 and name C    ) (segid AAAA and resid    3 and name N    )    1.0  65.4   120.7 2
        """.replace(
            NOQA_E501, ""
        ),
    ]

    for expected, reported in zip(EXPECTED, result):
        expected = [elem.strip() for elem in expected.split()]
        reported = [elem.strip() for elem in reported.split()]

        assert reported == expected


def test_get_single_atom_selection_dihderal_restraint():
    dihedrals_path = path_in_test_data(__file__, "test_2_dihedrals.tbl")

    dihdedral_restraints = open(dihedrals_path, "r").read()

    xplor_basic_restraints = _dihedral_restraints.ignore(XPLOR_COMMENT).parseString(
        dihdedral_restraints
    )

    restraint = xplor_basic_restraints[0].get("atoms_1")[0]

    selection_expressions = _get_selection_expressions_from_selection(restraint, "")

    assert len(selection_expressions) == 1

    assert str(selection_expressions[0]) == "[[segid : AAAA, resid : 1, atom : C]]"



from enum import Enum


class StereoAssignmentHandling(Enum):
    ALL_AMBIGUOUS = "all-ambiguous"
    AS_ASSIGNED = "as-assigned"
    AUTO = "auto"

    def __str__(self):
        return self._name_.lower().replace("_", "-")



from pathlib import Path

import typer
from typer import Option

from nef_pipelines.lib.nef_lib import read_or_create_entry_exit_error_on_bad_file
from nef_pipelines.lib.sequence_lib import frame_to_chains
from nef_pipelines.lib.util import STDIN
from nef_pipelines.tools.chains import chains_app

app = typer.Typer()

# TODO: it would be nice to put the chains with the first molecular system frame


# noinspection PyUnusedLocal
@chains_app.command()
def list(
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="input to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    comment: bool = Option(False, "-c", "--comment", help="prepend comment to chains"),
    verbose: bool = Option(False, "-v", "--verbose", help="print verbose info"),
    stream: bool = Option(False, "-s", "--stream", help="stream file after comment"),
):
    """- list the chains in the molecular systems"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    sequence_frames = entry.get_saveframes_by_category("nef_molecular_system")

    chains = []
    if len(sequence_frames) > 0:
        chains = frame_to_chains(sequence_frames[0])

    result = " ".join(chains)
    chains = "chain" if len(chains) == 1 else "chains"

    verbose = f"{len(result)} {chains}: " if verbose else ""

    comment = "# " if comment else ""

    print(f"{comment}{verbose}{result}")

    if stream:
        print(entry)



import typer

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xplor.importers.distances import distances

app = typer.Typer()
app.command()(distances)


# noinspection PyUnusedLocal
def test_2_distances():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    dihedrals_path = path_in_test_data(__file__, "test_2_distances.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [dihedrals_path]
    result = run_and_report(app, args, input=nef_sequence)

    EXPECTED = """\
        save_nef_distance_restraint_list_test_2_distances                               # noqa: E501
            _nef_distance_restraint_list.sf_category     nef_distance_restraint_list
            _nef_distance_restraint_list.sf_framecode    nef_distance_restraint_list_test_2_distances
            _nef_distance_restraint_list.potential_type  undefined

            loop_
                _nef_distance_restraint.index
                _nef_distance_restraint.restraint_id
                _nef_distance_restraint.restraint_combination_id
                _nef_distance_restraint.chain_code_1
                _nef_distance_restraint.sequence_code_1
                _nef_distance_restraint.residue_name_1
                _nef_distance_restraint.atom_name_1
                _nef_distance_restraint.chain_code_2
                _nef_distance_restraint.sequence_code_2
                _nef_distance_restraint.residue_name_2
                _nef_distance_restraint.atom_name_2
                _nef_distance_restraint.weight
                _nef_distance_restraint.target_value
                _nef_distance_restraint.lower_limit
                _nef_distance_restraint.upper_limit
                1   1   .   AAAA   1   .   HN   AAAA   2   .   HDA#   1.0   2.6   0.6   5.6
                2   2   .   AAAA   2   .   HN   AAAA   3   .   HD1#   1.0   4.5   1.5   6.2

           stop_
        save_
    """.replace(
        NOQA_E501, ""
    )

    result = isolate_frame(
        result.stdout, "nef_distance_restraint_list_test_2_distances"
    )

    assert_lines_match(EXPECTED, result)


def test_2_distances_bad():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    distances_path = path_in_test_data(__file__, "test_2_distances_bad.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [distances_path]
    result = run_and_report(app, args, input=nef_sequence, expected_exit_code=1)

    assert "ERROR" in result.stdout
    assert "failed to read distance restraints" in result.stdout


def test_2_dihstances_no_segids():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    distances_path = path_in_test_data(__file__, "test_2_distances_no_segids.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [distances_path, "--chains", "AAAA"]
    result = run_and_report(app, args, input=nef_sequence)

    EXPECTED = """\
        save_nef_distance_restraint_list_test_2_distances_no_segids                               # noqa: E501
            _nef_distance_restraint_list.sf_category     nef_distance_restraint_list
            _nef_distance_restraint_list.sf_framecode    nef_distance_restraint_list_test_2_distances_no_segids
            _nef_distance_restraint_list.potential_type  undefined

            loop_
                _nef_distance_restraint.index
                _nef_distance_restraint.restraint_id
                _nef_distance_restraint.restraint_combination_id
                _nef_distance_restraint.chain_code_1
                _nef_distance_restraint.sequence_code_1
                _nef_distance_restraint.residue_name_1
                _nef_distance_restraint.atom_name_1
                _nef_distance_restraint.chain_code_2
                _nef_distance_restraint.sequence_code_2
                _nef_distance_restraint.residue_name_2
                _nef_distance_restraint.atom_name_2
                _nef_distance_restraint.weight
                _nef_distance_restraint.target_value
                _nef_distance_restraint.lower_limit
                _nef_distance_restraint.upper_limit
                1   1   .   AAAA   1   .   HN   AAAA   2   .   HDA#   1.0   2.6   0.6   5.6
                2   2   .   AAAA   2   .   HN   AAAA   3   .   HD1#   1.0   4.5   1.5   6.2

           stop_
        save_
    """.replace(
        NOQA_E501, ""
    )

    result = isolate_frame(
        result.stdout, "nef_distance_restraint_list_test_2_distances_no_segids"
    )

    assert_lines_match(EXPECTED, result)


def test_2_distances_overriding_segids():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    distances_path = path_in_test_data(__file__, "test_2_distances.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [distances_path, "--chains", "BBBB", "--use-chains"]
    result = run_and_report(app, args, input=nef_sequence)

    EXPECTED = """\
        save_nef_distance_restraint_list_test_2_distances                               # noqa: E501
            _nef_distance_restraint_list.sf_category     nef_distance_restraint_list
            _nef_distance_restraint_list.sf_framecode    nef_distance_restraint_list_test_2_distances
            _nef_distance_restraint_list.potential_type  undefined

            loop_
                _nef_distance_restraint.index
                _nef_distance_restraint.restraint_id
                _nef_distance_restraint.restraint_combination_id
                _nef_distance_restraint.chain_code_1
                _nef_distance_restraint.sequence_code_1
                _nef_distance_restraint.residue_name_1
                _nef_distance_restraint.atom_name_1
                _nef_distance_restraint.chain_code_2
                _nef_distance_restraint.sequence_code_2
                _nef_distance_restraint.residue_name_2
                _nef_distance_restraint.atom_name_2
                _nef_distance_restraint.weight
                _nef_distance_restraint.target_value
                _nef_distance_restraint.lower_limit
                _nef_distance_restraint.upper_limit
                1   1   .   BBBB   1   .   HN   BBBB   2   .   HDA#   1.0   2.6   0.6   5.6
                2   2   .   BBBB   2   .   HN   BBBB   3   .   HD1#   1.0   4.5   1.5   6.2

           stop_
        save_
    """.replace(
        NOQA_E501, ""
    )

    result = isolate_frame(
        result.stdout, "nef_distance_restraint_list_test_2_distances"
    )

    assert_lines_match(EXPECTED, result)



import sys
from pathlib import Path
from typing import Dict, List, Tuple

import typer
from fyeah import f
from pynmrstar import Entry, Saveframe
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    read_entry_from_stdin_or_exit,
    select_frames_by_name,
)
from nef_pipelines.lib.sequence_lib import (
    TRANSLATIONS_3_1_PROTEIN,
    get_residue_name_from_lookup,
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.structures import ShiftData
from nef_pipelines.lib.util import STDOUT, exit_error, flatten, is_int
from nef_pipelines.transcoders.shifty import export_app

FILE_NAME_TEMPLATE = "{nef_entry_id}_{chain_code}.shifty"
OUTPUT_FILE_HELP = (
    f"file name to output to [default is {FILE_NAME_TEMPLATE}; - gives stdout]"
)
FILE_NAME_TEMPLATE_HELP = f"file name template default is [{FILE_NAME_TEMPLATE}]"

SHIFTY_HEADINGS_LOOKUP = (
    ("HA", ("HA1", "HA")),
    ("HN", ("H", "HN")),  # TODO: this is a hack awaiting the translations module
    ("N15", ("N",)),
    ("CA", ("CA",)),
    ("CB", ("CB",)),
    ("CO", ("C",)),
)
SHIFTY_ATOMS = [shifty_atom for shifty_atom, _ in SHIFTY_HEADINGS_LOOKUP]
SHIFTY_LOOKUP_ATOMS = flatten(
    [lookup_atom for _, lookup_atom in SHIFTY_HEADINGS_LOOKUP]
)
SHIFTY_EXTRA_HEADINGS = "#NUM AA".split()
SHIFTY_HEADINGS = [*SHIFTY_EXTRA_HEADINGS, *SHIFTY_ATOMS]


# noinspection PyUnusedLocal
@export_app.command()
def shifts(
    shift_frames: List[str] = typer.Option(
        [],
        "-f",
        "--frame",
        help="selector for the shift restraint frames to use, can be called multiple times and include wild cards",
    ),
    chain_code: str = typer.Option(
        None,
        "-c",
        "--chain",
        help="chain to export [defaults to the first chain if only one present]",
        metavar="<CHAIN-CODE>",
    ),
    file_name_template: str = typer.Option(
        FILE_NAME_TEMPLATE, help=FILE_NAME_TEMPLATE_HELP
    ),
    infill: bool = typer.Option(
        True, help="infill missing residues with values set to 0.0"
    ),
    output_file: str = typer.Argument(
        None,
        help=OUTPUT_FILE_HELP,
        metavar="<SHIFTY-SHIFT-FILE-NAME>",
    ),
):
    """- convert nef chemical shifts to shifty"""

    if len(shift_frames) == 0:
        shift_frames = ["*"]

    entry = read_entry_from_stdin_or_exit()
    nef_entry_id = entry.entry_id  # noqa: F841

    sequence = sequence_from_entry_or_exit(entry) if infill else None
    sequence_lookup = sequence_to_residue_name_lookup(sequence) if sequence else None

    if chain_code is None:
        chain_codes = set([residue.chain_code for residue in sequence])
        assigned_chain_codes = set(
            [chain_code for chain_code in chain_codes if chain_code not in ["@", "#"]]
        )

        if len(assigned_chain_codes) == 1:
            chain_code = assigned_chain_codes.pop()
        else:
            msg = f"""you must chhose a single assigned chain i got {', '.join(assigned_chain_codes)}"""
            exit_error(msg)

    if output_file is None:
        output_file = f(file_name_template)

    output_file = Path(output_file) if output_file != "-" else STDOUT

    entry = pipe(entry, shift_frames, chain_code, infill, sequence_lookup, output_file)

    if not (sys.stdout.isatty() or output_file == STDOUT):
        print(entry)


def pipe(
    entry: Entry,
    shift_frames: List[Saveframe],
    selected_chain_code: str,
    infill: bool,
    residue_name_lookup: Dict[Tuple[str, str], str],
    output_file: Path,
) -> Entry:

    frames = entry.get_saveframes_by_category("nef_chemical_shift_list")

    frames = select_frames_by_name(frames, shift_frames)

    if len(frames) == 0:
        exit_error("no shift frames selected")

    shifts = nef_frames_to_shifts(frames)

    selected_chain_code = _get_chain_code_or_exit(shifts, selected_chain_code, entry)

    lines = {}

    for shift in shifts:

        chain_code = shift.atom.residue.chain_code
        sequence_code = shift.atom.residue.sequence_code
        atom_name = shift.atom.atom_name
        residue_name = shift.atom.residue.residue_name

        if _is_pseudo_residue(chain_code, sequence_code):
            continue

        if residue_name.upper() not in TRANSLATIONS_3_1_PROTEIN:
            continue

        if atom_name not in SHIFTY_LOOKUP_ATOMS:
            continue

        if chain_code != selected_chain_code:
            continue

        if not is_int(sequence_code):
            continue

        sequence_code = int(sequence_code)
        residue_name_1let = TRANSLATIONS_3_1_PROTEIN[residue_name]
        value = shift.value

        key = sequence_code, residue_name_1let
        lines.setdefault(key, {})[atom_name] = value

    if infill:
        lines = _infill_sequence_exit(lines, chain_code, residue_name_lookup)

    table = [[*SHIFTY_EXTRA_HEADINGS, *SHIFTY_ATOMS]]

    for (residue_number, residue_type), shifts in lines.items():

        row = [residue_number, residue_type]
        table.append(row)

        shift_values = []
        for shift_name, lookup_atom_names in SHIFTY_HEADINGS_LOOKUP:
            shift_value = 0.0
            for lookup_atom_name in lookup_atom_names:
                if lookup_atom_name in shifts:
                    shift_value = shifts[lookup_atom_name]

            # if we use floats the trailing zeros get removed
            # shift_values.append(shift_value)
            shift_values.append(format(shift_value, "7.3f"))
        row.extend(shift_values)

    file_h = sys.stdout if output_file == STDOUT else open(output_file, "w")

    COL_ALIGN = [
        "right",
    ] * len(SHIFTY_HEADINGS)
    COL_ALIGN[0] = "left"
    COL_ALIGN[1] = "left"

    print(tabulate(table, tablefmt="plain", colalign=COL_ALIGN), file=file_h)

    if output_file != STDOUT:
        file_h.close()

    return entry


def _infill_sequence_exit(lines, chain_code, residue_name_lookup):
    sequence_codes = [sequence_code for sequence_code, _ in lines.keys()]
    max_sequence_code, min_sequence_code = _sequence_code_range_or_exit_on_non_ints(
        chain_code, sequence_codes
    )

    for sequence_code in range(min_sequence_code, max_sequence_code + 1):
        residue_name = get_residue_name_from_lookup(
            chain_code, sequence_code, residue_name_lookup
        )
        residue_name_1_let = TRANSLATIONS_3_1_PROTEIN[residue_name]
        key = sequence_code, residue_name_1_let
        if key not in lines:
            lines[key] = {atom_name: 0.0 for atom_name in SHIFTY_LOOKUP_ATOMS}
    lines = dict(sorted(lines.items()))
    return lines


def _sequence_code_range_or_exit_on_non_ints(chain_code, sequence_codes):
    int_sequence_codes = [
        sequence_code for sequence_code in sequence_codes if is_int(sequence_code)
    ]
    min_sequence_code = min(int_sequence_codes)
    max_sequence_code = max(int_sequence_codes)
    is_int_list = [is_int(sequence_code) for sequence_code in sequence_codes]
    all_ints = all(is_int_list)
    if not all_ints:
        msg = f"""\
            the sequence codes for chain {chain_code} aren't all integers
            the sequence codes must be integers to infill missing residues
            the first non-integer sequence code is {min_sequence_code + sequence_codes[is_int_list.index(False)]}
        """
        exit_error(msg)
    return max_sequence_code, min_sequence_code


def _is_pseudo_residue(chain_code, sequence_code):
    return (
        chain_code.startswith("@-")
        or not is_int(sequence_code)
        and sequence_code.startswith("@")
    )


def _filter_headings_by_pseudoatoms(base_headers, pseudo_residues):
    atom_names = set()
    for pseudo_residue in pseudo_residues.values():
        for pseudo_atom in pseudo_residue.values():
            offset = (
                f"-{pseudo_atom.negative_offset}"
                if pseudo_atom.negative_offset == 1
                else ""
            )
            atom_names.add(f"{pseudo_atom.atom_name.upper()}{offset}")
    headers = []
    for header in base_headers:
        if header == "" or header in atom_names:
            headers.append(header)
    return headers


def _get_chain_code_or_exit(
    read_shifts: ShiftData, selected_chain_code: str, entry: Entry
):

    chain_codes = set([shift.atom.residue.chain_code for shift in read_shifts])

    if selected_chain_code is None:

        if len(chain_codes) > 1:
            msg = f"""\
                no chain code selected and ambiguous chain codes available [{' '.join(chain_codes)}]
            """
            exit_error(msg)
        else:
            selected_chain_code = list(chain_codes)[0]

    if selected_chain_code not in chain_codes:
        msg = f"""\
            the chain code you selected [{selected_chain_code}] isn't in the chain codes [{' '.join(chain_codes)}]
            in the entry {entry.entry_id}
        """
        exit_error(msg)

    return selected_chain_code



import typer

from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:
    nef_app.app.add_typer(
        app,
        name="xcamshift",
        help="- write xcamshift for xplor [shifts]",
    )

    app.add_typer(
        export_app,
        name="export",
        help="- export xcamhift for xplor [shifts]",
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.xcamshift.exporters.shifts  # noqa: F401



import inspect
import sys
from fnmatch import fnmatch
from typing import List

import typer
from pynmrstar import Entry

from nef_pipelines.lib.util import exit_error, running_in_pycharm
from nef_pipelines.tools.frames import frames_app

UNDERSCORE = "_"

parser = None


# noinspection PyUnusedLocal
@frames_app.command()
def delete(
    use_categories: bool = typer.Option(
        False,
        "-c",
        "--category",
        help="if selected use the category of the frame to select it for deletion rather than it name",
    ),
    exact: bool = typer.Option(
        False, "-e", "--exact", help="don't treat name as a wild card"
    ),
    selectors: List[str] = typer.Argument(
        ...,
        help="a list of frames to delete by type or name,  names can be wildcards, names have lead _'s removed and "
        "surrounding back quotes `  removed",
    ),
):
    """- delete frames in the current input by type or name"""

    entry = _create_entry_from_stdin_or_exit(current_function())

    to_delete = []
    for name in selectors:
        for frame in entry:
            frame_full_name = frame.name
            frame_category = frame.category
            frame_name = frame_full_name[len(frame_category) :].lstrip("_").strip("`")

            for selector in selectors:
                if not exact:
                    selector = f"*{selector}*"

                if use_categories:
                    if fnmatch(frame_category, selector):
                        to_delete.append(frame)
                else:
                    if fnmatch(frame_name, selector):
                        to_delete.append(frame)

    entry.remove_saveframe(to_delete)

    print(entry)


def current_function():

    return inspect.stack()[1][3]


def calling_function():

    return inspect.stack()[2][3]


# TODO: This should be a library function
def _create_entry_from_stdin_or_exit(command_name: str):

    try:

        if sys.stdin.isatty():
            exit_error(
                f"the command {command_name} reads from stdin and there is no stream..."
            )

        if running_in_pycharm():
            exit_error("you can't build read fron stdin in pycharm...")

        lines = list(iter(sys.stdin))

        if len(lines) == 0:
            exit_error(
                f"the command {command_name} reads from stdin and the stream is empty..."
            )

        entry = Entry.from_string("".join(lines))

    except Exception as e:
        exit_error(f"failed to read nef entry from stdin because {e}", e)

    return entry



import pytest
from frozendict import frozendict

from nef_pipelines.lib.test_lib import path_in_test_data
from nef_pipelines.transcoders.rcsb.rcsb_lib import (
    PdbSecondaryStructureType,
    parse_cif,
    parse_pdb, guess_cif_or_pdb, RCSBFileType,
)

# # test=PdbxReader()
#
# def test_short_cif():
#     short_file_name = "1l2y_short.cif"
#     file_path = path_in_test_data(__file__, short_file_name)
#     lines = open(file_path)
#
#     structure = parse_cif(lines, short_file_name)
#     assert len(structure.sequences) == 1
#     assert len(structure.secondary_structure) == 1
#     assert len(structure.models) == 2
#
#     model_0 = structure.models[0]
#     model_1 = structure.models[1]
#     assert len(model_0.chains) == len(model_1.chains) == 1
#
#     model_0_chain_A = model_0.chains['A']
#     model_1_chain_A = model_1.chains['A']
#     assert len(model_0_chain_A.residues) == len(model_1_chain_A.residues) == 4
#
#     assert model_0_chain_A.residues[0] != model_1_chain_A.residues[0]
#
#     assert list(structure.sequences.keys()) == ['1',]
#     assert structure.sequences['1'].id == '1'
#     assert structure.sequences['1'].residues == ['ASN', 'LEU', 'TYR', 'ILE']
#     assert structure.sequences['1'].structure is structure
#
#     secondary_structure =structure.secondary_structure['A'][0]
#     assert secondary_structure.chain_code == 'A'
#     assert secondary_structure.start_sequence_code ==  1
#     assert secondary_structure.end_sequence_code == 4
#     assert secondary_structure.alternative_location == None
#     assert secondary_structure.secondary_structure_type == PdbSecondaryStructureType.HELIX_RIGHT_HANDED_ALPHA
#     assert secondary_structure.structure == structure
#
#     EXPECTED_ATOMS = {
#         frozendict({'model': 1, 'chain_code': 'A', 'sequence_code': 1, 'residue_name': 'ASN', 'atom_number': 0}): {
#             'name': 'N',
#             'x': -8.901,
#             'y': 4.127,
#             'z': -0.555,
#             'element': 'N',
#         },
#
#         frozendict({'model': 1, 'chain_code': 'A', 'sequence_code': 3,'residue_name': 'TYR', 'atom_number': 19}): {
#             'name': 'HE2',
#             'x': 0.033,
#             'y': 4.952,
#             'z': 4.233,
#             'element': 'H',
#         },
#
#         frozendict({'model': 2, 'chain_code': 'A', 'sequence_code': 2, 'residue_name': 'LEU', 'atom_number': 4}): {
#             'name': 'CB',
#             'x': -4.073,
#             'y':  4.863,
#             'z': -3.128,
#             'element': 'C',
#         },
#
#         frozendict({'model': 2, 'chain_code': 'A', 'sequence_code': 4, 'residue_name': 'ILE', 'atom_number': 18}): {
#             'name': 'HD13',
#             'x': -9.973,
#             'y':  1.875,
#             'z':  1.360,
#             'element': 'H',
#         },
#     }
#
#     for key_values, data_values in EXPECTED_ATOMS.items():
#         model = structure.models[key_values['model']-1]
#         chain = model.chains[key_values['chain_code']]
#         residue = chain.residues[key_values['sequence_code']-1]
#         atom = residue.atoms[key_values['atom_number']]
#
#         assert model.serial == key_values['model']
#         assert chain.chain_code == key_values['chain_code']
#         assert residue.chain.chain_code == key_values['chain_code']
#         assert atom.residue.name == key_values['residue_name']
#         for key, value in data_values.items():
#             assert getattr(atom, key) == value


FILE_TYPE_TO_READER  =  {
    'cif': parse_cif,
    'pdb': parse_pdb
}
@pytest.mark.parametrize(
    "file_name",
    [
        ("1l2y_short.pdb"),
        ("1l2y_short.cif"),
    ],
)
def test_short(file_name):


    file_path = path_in_test_data(__file__, file_name)
    lines = open(file_path)

    file_type = file_name.split('.')[-1]
    reader = FILE_TYPE_TO_READER[file_type]
    structure = reader(lines, file_name)
    assert len(structure.sequences) == 1
    assert len(structure.secondary_structure) == 1
    assert len(structure.models) == 2

    model_0 = structure.models[0]
    model_1 = structure.models[1]
    assert len(model_0.chains) == len(model_1.chains) == 1

    model_0_chain_A = model_0.chains["A"]
    model_1_chain_A = model_1.chains["A"]
    assert len(model_0_chain_A.residues) == len(model_1_chain_A.residues) == 4

    assert model_0_chain_A.residues[0] != model_1_chain_A.residues[0]

    assert list(structure.sequences.keys()) == [
        "1",
    ]
    assert structure.sequences["1"].id == "1"
    assert structure.sequences["1"].residues == ["ASN", "LEU", "TYR", "ILE"]
    assert structure.sequences["1"].structure is structure

    secondary_structure = structure.secondary_structure["A"][0]
    assert secondary_structure.chain_code == "A"
    assert secondary_structure.start_sequence_code == 1
    assert secondary_structure.end_sequence_code == 4
    assert secondary_structure.alternative_location is None
    assert (
        secondary_structure.secondary_structure_type
        == PdbSecondaryStructureType.HELIX_RIGHT_HANDED_ALPHA
    )
    assert secondary_structure.structure == structure

    EXPECTED_ATOMS = {
        frozendict(
            {
                "model": 1,
                "chain_code": "A",
                "sequence_code": 1,
                "residue_name": "ASN",
                "atom_number": 0,
            }
        ): {
            "atom_name": "N",
            "x": -8.901,
            "y": 4.127,
            "z": -0.555,
            "element": "N",
        },
        frozendict(
            {
                "model": 1,
                "chain_code": "A",
                "sequence_code": 3,
                "residue_name": "TYR",
                "atom_number": 19,
            }
        ): {
            "atom_name": "HE2",
            "x": 0.033,
            "y": 4.952,
            "z": 4.233,
            "element": "H",
        },
        frozendict(
            {
                "model": 2,
                "chain_code": "A",
                "sequence_code": 2,
                "residue_name": "LEU",
                "atom_number": 4,
            }
        ): {
            "atom_name": "CB",
            "x": -4.073,
            "y": 4.863,
            "z": -3.128,
            "element": "C",
        },
        frozendict(
            {
                "model": 2,
                "chain_code": "A",
                "sequence_code": 4,
                "residue_name": "ILE",
                "atom_number": 18,
            }
        ): {
            "atom_name": "HD13",
            "x": -9.973,
            "y": 1.875,
            "z": 1.360,
            "element": "H",
        },
    }

    for key_values, data_values in EXPECTED_ATOMS.items():
        model = structure.models[key_values["model"] - 1]
        chain = model.chains[key_values["chain_code"]]
        residue = chain.residues[key_values["sequence_code"] - 1]
        atom = residue.atoms[key_values["atom_number"]]

        assert model.serial == key_values["model"]
        assert chain.chain_code == key_values["chain_code"]
        assert residue.chain.chain_code == key_values["chain_code"]
        assert atom.residue.residue_name == key_values["residue_name"]
        for key, value in data_values.items():
            assert getattr(atom, key) == value

def test_file_type_from_extension():
    assert guess_cif_or_pdb('', 'test.mmcif') == RCSBFileType.CIF
    assert guess_cif_or_pdb('', 'test.pdb') == RCSBFileType.PDB


@pytest.mark.parametrize(
    'file_name',
    [
        ("1l2y_short.pdb"),
        ("1l2y_short.cif"),
    ],
)
def test_file_type_from_data(file_name):

    file_type_str = file_name.split('.')[-1]

    file_path = path_in_test_data(__file__, file_name)
    file_data = open(file_path).readlines()
    assert guess_cif_or_pdb(file_data) == RCSBFileType[file_type_str.upper()]




import functools
import operator
import re
from collections import Counter
from dataclasses import dataclass, field
from enum import IntEnum
from textwrap import dedent
from typing import Callable, List, Optional, TextIO, Tuple, Union

from tabulate import tabulate

from nef_pipelines.lib.sequence_lib import (
    MoleculeTypes,
    make_chunked_sequence_1let,
    offset_chain_residues,
    sequence_3let_to_sequence_residues,
    translate_1_to_3,
)
from nef_pipelines.lib.structures import (
    AtomLabel,
    LineInfo,
    PeakAxis,
    PeakList,
    PeakListData,
    PeakValues,
    SequenceResidue,
    ShiftData,
    ShiftList,
)
from nef_pipelines.lib.util import exit_error, is_int


class PEAK_TYPES(IntEnum):
    PEAK = 1
    NOISE = 2
    SINC_WIGGLE = 3


# known field types
DATA = "DATA"
REMARK = "REMARK"
VALUES = "__VALUES__"
SEQUENCE = "SEQUENCE"
FORMAT = "FORMAT"
VARS = "VARS"
COMMENT = "#"
NULLSTRING = "NULLSTRING"
NULLVALUE = "NULLVALUE"

NMRPIPE_PEAK_EXPECTED_FIELDS = "INDEX X_AXIS XW XW_HZ ASS CLUSTID MEMCNT".split()
NMRPIPE_SHIFTS_EXPECTED_FIELDS = "RESID RESNAME ATOMNAME SHIFT".split()


@dataclass(frozen=True)
class DbRecord:
    index: int
    type: str
    values: Tuple[Union[int, str, float]]
    line_info: LineInfo = None


@dataclass
class DbFile:
    name: str = "unknown"
    records: List[DbRecord] = field(default_factory=list)


def _raise_data_before_format(line_info):
    msg = f"""\
            bad nmrpipe db file, data seen before VAR and FORMAT
            file: {line_info.file_name}'
            line no: {line_info.line_no}
            line: {line_info.line}
            """
    msg = dedent(msg)
    raise DataBeforeFormat(msg)


def read_db_file_records(file_h: TextIO, file_name: str = "unknown") -> DbFile:

    """
    Read from NmrPipe (NIH) tab/gdb-file
    Args:
        file_h (TextIO): a file like object
        file_name (str): the name of the file being read (for debugging)

    Returns DbFile:
        a list of all the records in the fiule
    """

    records: List[DbRecord] = []
    column_names = None
    column_formats = None
    record_count = Counter()
    in_header = True

    for line_index, line in enumerate(file_h):

        line_info = LineInfo(file_name, line_index + 1, line)
        line = line.strip()

        raw_fields = line.strip().split()

        if len(line) == 0:
            continue

        record_type = raw_fields[0]
        record_count[record_type] += 1

        fields = raw_fields

        handled = False

        if in_header:
            if record_type == "VARS":

                column_names = fields[1:]
                records.append(
                    DbRecord(
                        record_count[record_type], record_type, column_names, line_info
                    )
                )
                handled = True

                if record_count[record_type] != 1:
                    _raise_multiple(record_type, line_info)

            if record_type == "FORMAT":
                column_formats = _formats_to_constructors(fields, line_info)

                _check_var_and_format_count_raise_if_bad(
                    column_names, column_formats, line_info
                )

                records.append(
                    DbRecord(
                        record_count[record_type], record_type, fields[1:], line_info
                    )
                )
                in_header = False
                continue

            if record_type in ("REMARK", "#"):
                records.append(
                    DbRecord(record_count[record_type], record_type, line, line_info)
                )
                handled = True

            if record_type == DATA:
                records.append(
                    DbRecord(
                        record_count[record_type], record_type, fields[1:], line_info
                    )
                )
                handled = True

            if not handled:
                _raise_data_before_format(line_info)

        if not in_header:

            if record_type in (VARS, FORMAT):
                _raise_multiple(record_type, line_info)

            if column_names and column_formats:
                record_count["__VALUES__"] += 1

                del record_count[record_type]

                values = _build_values_or_raise(
                    column_formats, column_names, fields, line_info
                )

                record = DbRecord(
                    record_count["__VALUES__"], "__VALUES__", values, line_info
                )
                records.append(record)

                handled = True

        if not handled:
            records.append(
                DbRecord(record_count[record_type], record_type, fields[1:], line_info)
            )

    return DbFile(file_name, records)


def _find_nth(haystack, needle, n):
    # https://stackoverflow.com/questions/1883980/find-the-nth-occurrence-of-substring-in-a-string/41626399#41626399
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start + len(needle))
        n -= 1
    return start


def _build_values_or_raise(column_formats, column_names, fields, line_info):
    columns_formats = _check_column_count_raise_if_bad(
        column_formats, column_names, line_info
    )
    result = []
    field_count = Counter()
    for column_no, (raw_field, constructor) in enumerate(zip(fields, columns_formats)):
        try:
            field_count[raw_field] += 1
            value = constructor(raw_field)

        except Exception:
            absolute_column = _find_nth(
                line_info.line, raw_field, field_count[raw_field]
            )
            msg = f"""
                    Couldn't convert {raw_field} to type {_constructor_to_name(constructor)}
                    file: {line_info.file_name}
                    line no: {line_info.line_no}
                    column: {column_no + 1}
                    line: {line_info.line.rstrip()}
                          {' ' * absolute_column + '^'}
                """
            msg = dedent(msg)
            raise BadFieldFormat(msg)
        result.append(value)
    return result


def _raise_multiple(format_str, line_info):
    msg = f"""\
                bad NMRPipe db file, multiple {format_str} statements found
                file: {line_info.file_name}
                line no: {line_info.line_no}
                line: {line_info.line}
                """
    msg = dedent(msg)

    if format_str == "VARS":
        raise MultipleVars(msg)
    elif format_str == "FORMAT":
        raise MultipleFormat(msg)


def _check_var_and_format_count_raise_if_bad(column_names, column_formats, line_info):
    if column_names is None:
        msg = f"""\
               no column names defined by a VARS line when FORMAT line read
               file: {line_info.file_name}
               line no: {line_info.line_no}
               line: {line_info.line}"""
        msg = dedent(msg)
        raise NoVarsLine(msg)

    num_formats = len(column_names)
    num_column_names = len(column_formats)
    if num_formats != num_column_names:
        msg = f"""\
                  number of column names and formats must agree
                  got {num_column_names} column names and {num_formats} formats
                  file: {line_info.file_name}
                  line no: {line_info.line_no}
                  line: {line_info.line}
               """
        msg = dedent(msg)

        raise WrongColumnCount(msg)


def _check_column_count_raise_if_bad(column_formats, column_names, line_info):
    raw_fields = line_info.line.split()
    num_fields = len(raw_fields)
    num_columns = len(column_formats)

    missing_fields = ["*"] * abs(num_fields - num_columns)
    raw_fields = [*raw_fields, *missing_fields]

    if num_fields != num_columns:
        column_formats = _constructor_names(column_formats)
        tab = [
            column_names,
            column_formats,
            raw_fields,
        ]
        tabulated = tabulate(tab, tablefmt="plain")
        msg = f"""\
                number fields ({num_fields + 1}) doesn't not match number of columns ({num_columns + 1})

                expected
                %s

                missing fields marked with *

                file: {line_info.file_name}
                line no: {line_info.line_no}
                line: {line_info.line}

            """
        msg = dedent(msg)
        msg = msg % tabulated
        raise WrongColumnCount(msg)
    return column_formats


def _formats_to_constructors(formats, line_info):
    result = []

    field_counter = Counter()
    for column_index, field_format in enumerate(formats[1:]):
        field_counter[field_format] += 1
        field_format = field_format.strip()
        field_format = field_format[-1]

        if field_format == "d":
            result.append(int)
        elif field_format == "f":
            result.append(float)
        elif field_format == "s":
            result.append(str)
        elif field_format == "e":
            result.append(float)
        else:
            format_column = _find_nth(
                line_info.line, field_format, field_counter[field_format]
            )
            msg = f"""
                unexpected format {field_format} at index {column_index+1}, expected formats are:
                s, d, e, f (string, integer, scientific(float), float)

                file: {line_info.file_name}
                line no: {line_info.line_no}
                line: {line_info.line}
                      {' ' * format_column + '^'}

            """
            raise BadFieldFormat(msg)
    return result


OptionDbRecordPredicate = Optional[Callable[[DbRecord], bool]]


def select_records(
    gdb: DbFile, record_type: str, predicate: OptionDbRecordPredicate = None
) -> List[DbRecord]:
    """
    Select records from a gdb file by type and predicate
    Args:
        gdb (DbFile): gdb/tab file
        type (str): the type of the record #, REMARK, __VALUE__ etc
        predicate (OptionDbRecordPredicate): an optional test to apply to the record

    Returns List[DbRecord]:
        in the selected gdb/tab records
    """
    result = [record for record in gdb.records if record.type == record_type]
    if predicate:
        result = [record for record in result if predicate(record)]
    return result


def select_data_records(gdb: DbFile, type: str) -> List[DbRecord]:
    """
    Select data records from the db file that have the specific type

    :param gdb:  a parsed GBD files
    :param type: the type of data record to select
    :return: a list of matching DbRecords
    """
    return select_records(gdb, "DATA", predicate=lambda rec: rec.values[0] == type)


def gdb_to_3let_sequence(
    gdb: DbFile, molecule_type: MoleculeTypes = MoleculeTypes.PROTEIN
) -> List[SequenceResidue]:
    """
    read sequence records from a gdb file and convert them to a list of sequence residues
    it is assumed that residues start from 1 and are in chain A
    Args:
        gdb (DbFile): the source db/tab file records
        molecule_type (MoleculeTypes): the type of the molecule to translate

    Returns List[SequenceResidue]:
        a list of sequence residues
    """
    sequence_records = select_data_records(gdb, "SEQUENCE")

    sequence = [record.values[1:] for record in sequence_records]

    flattened_sequence = functools.reduce(operator.iconcat, sequence, [])
    sequence_string = "".join(flattened_sequence)

    return translate_1_to_3(sequence_string, molecule_type=molecule_type)


def gdb_to_chain_start(gdb_file: DbFile) -> int:
    """
    Find a first resid record in a GbdFile and return it or return 1
    :param gdb_file: a GBD file
    :return: a chain start
    """

    chain_start_records = select_data_records(gdb_file, "FIRST_RESID")

    if len(chain_start_records) == 0:
        chain_start = 1
    else:
        chain_start = int(chain_start_records[-1].values[1])

    return chain_start


def gdb_to_sequence(gdb_file: DbFile, chain_code: str) -> List[SequenceResidue]:

    """
    Read the sequence from a GDB file, note if a FIRST_RESID record is found the sequence
    is offset correctly...
    :param gdb_file:  a GDB file
    :param chain_code: the chain code to give the sequence
    :return: a List of sequence residues
    """
    chain_start = gdb_to_chain_start(gdb_file)

    sequence_3let = gdb_to_3let_sequence(gdb_file)

    sequence_residues = sequence_3let_to_sequence_residues(
        sequence_3let, chain_code=chain_code
    )

    sequence_residues = offset_chain_residues(
        sequence_residues, {chain_code: chain_start - 1}
    )

    return sequence_residues


def _assignments_to_atom_labels(assignments, dimensions, chain_code="A"):
    result = []

    for assignment in assignments:
        chain_code = chain_code

        residue_name = None
        len_assignment = len(assignment)
        if len_assignment > 0:
            residue_name = translate_1_to_3(assignment[0], unknown=".")[0]

        sequence_code = None
        if len_assignment > 1:
            raw_sequence_code = assignment[1]
            if is_int(raw_sequence_code):
                sequence_code = int(raw_sequence_code)

        atom_name = None
        if len_assignment > 2:
            atom_name = assignment[2]

        result.append(
            AtomLabel(
                SequenceResidue(chain_code, sequence_code, residue_name), atom_name
            )
        )

    len_result = len(result)
    if len_result < dimensions:
        for i in len_result - dimensions:
            result.append(AtomLabel(SequenceResidue(None, None, None), None))
    return result


# NMRPIPE VARS https://spin.niddk.nih.gov/NMRPipe/doc2new/
#
# INDEX = 'INDEX'  # REQUIRED      - The unique peak ID number.
# X_AXIS = 'X_AXIS'  # REQUIRED      - Peak position: in points in 1st dimension, from left of spectrum limit
# Y_AXIS = 'Y_AXIS'  # REQUIRED      - Peak position: in points in 2nd dimension, from bottom of spectrum limit
# DX = 'DX'  # NOT REQUIRED  - Estimate of the error in peak position due to random noise, in points.
# DY = 'DY'  # NOT REQUIRED  - Estimate of the error in peak position due to random noise, in points.
# X_PPM = 'X_PPM'  # NOT REQUIRED  - Peak position: in ppm in 1st dimension
# Y_PPM = 'Y_PPM'  # NOT REQUIRED  - Peak position: in ppm in 2nd dimension
# X_HZ = 'X_HZ'  # NOT REQUIRED  - Peak position: in Hz in 1st dimension
# Y_HZ = 'Y_HZ'  # NOT REQUIRED  - Peak position: in Hz in 2nd dimension
# XW = 'XW'  # REQUIRED      - Peak width: in points in 1st dimension
# YW = 'YW'  # REQUIRED      - Peak width: in points in 2nd dimension
# XW_HZ = 'XW_HZ'  # REQUIRED      - Peak width: in points in 1st dimension
# YW_HZ = 'YW_HZ'  # REQUIRED      - Peak width: in points in 2nd dimension
# X1 = 'X1'  # NOT REQUIRED  - Left border of peak in 1st dim, in points
# X3 = 'X3'  # NOT REQUIRED  - Right border of peak in 1st dim, in points
# Y1 = 'Y1'  # NOT REQUIRED  - Left border of peak in 2nd dim, in points
# Y3 = 'Y3'  # NOT REQUIRED  - Right border of peak in 2nd, in points
# HEIGHT = 'HEIGHT'  # NOT REQUIRED  - Peak height
# DHEIGHT = 'DHEIGHT'  # NOT REQUIRED  - Peak height error
# VOL = 'VOL'  # NOT REQUIRED  - Peak volume
# PCHI2 = 'PCHI2'  # NOT REQUIRED  - the Chi-square probability for the peak (i.e. probability due to the noise)
# TYPE = 'TYPE'  # NOT REQUIRED  - the peak classification; 1=Peak, 2=Random Noise, 3=Truncation artifact.
# ASS = 'ASS'  # REQUIRED      - Peak assignment
# CLUSTID = 'CLUSTID'  # REQUIRED      - Peak cluster id. Peaks with the same CLUSTID value are the overlapped.
# MEMCNT = 'MEMCNT'  # REQUIRED      - the total number of peaks which are in a given peak's cluster
# (i.e. peaks which have the same CLUSTID value)


def get_column_indices(gdb_file):
    return {column: index for index, column in enumerate(get_gdb_columns(gdb_file))}


def _mean(values):
    return sum(values) / len(values)


def _get_peak_list_dimension(gdb_file):

    return len(_get_axis_labels(gdb_file))


def _get_axis_labels(gdb_file):
    columns = get_gdb_columns(gdb_file)

    result = []
    for var in columns:
        if var.endswith("_AXIS"):
            result.append(var.split("_")[0])
    return result


def get_gdb_columns(gdb_file: DbFile) -> List[str]:
    return select_records(gdb_file, VARS)[0].values


def check_is_peak_file(gdb_file):
    columns = set(get_gdb_columns(gdb_file))
    expected_fields = set(NMRPIPE_PEAK_EXPECTED_FIELDS)

    return expected_fields.issubset(columns)


def check_is_shifts_file(gdb_file):
    columns = set(get_gdb_columns(gdb_file))
    expected_fields = set(NMRPIPE_SHIFTS_EXPECTED_FIELDS)

    return expected_fields.issubset(columns)


def read_peak_file(gdb_file, args):
    data = select_records(gdb_file, VALUES)

    dimensions = _get_peak_list_dimension(gdb_file)

    column_indices = get_column_indices(gdb_file)

    axis_labels = _get_axis_labels(gdb_file)

    spectrometer_frequencies = [[] for _ in range(dimensions)]

    raw_peaks = []
    for index, line in enumerate(data, start=1):

        peak = {}
        raw_peaks.append(peak)
        peak_type = line.values[column_indices["TYPE"]]
        if args.filter_noise and peak_type != PEAK_TYPES.PEAK:
            continue

        assignment = line.values[column_indices["ASS"]]
        assignments = assignment.split("-")
        assignments = _propagate_assignments(assignments)
        assignments = _assignments_to_atom_labels(assignments, dimensions)

        height = line.values[column_indices["HEIGHT"]]
        # height_error = line.values[column_indices["DHEIGHT"]]
        # height_percentage_error = height_error / height
        volume = line.values[column_indices["VOL"]]
        # volume_error = volume * height_percentage_error

        peak_values = PeakValues(
            serial=index, volume=volume, height=height, deleted=False, comment=""
        )
        peak["values"] = peak_values

        for i, dimension in enumerate("X Y Z A".split()[:dimensions]):

            shift = line.values[column_indices["%s_PPM" % dimension]]

            # point_error = line.values[column_indices["D%s" % dimension]]

            # point = line.values[column_indices["%s_AXIS" % dimension]]
            # shift_error = point_error / point * shift

            pos_hz = line.values[column_indices["%s_HZ" % dimension]]

            axis = PeakAxis(atom_labels=assignments[i], ppm=shift, merit=1)

            peak[i] = axis

            sf = pos_hz / shift

            spectrometer_frequencies[i].append(sf)

    spectrometer_frequencies = [
        _mean(frequencies) for frequencies in spectrometer_frequencies
    ]
    header_data = PeakListData(
        num_axis=dimensions,
        axis_labels=axis_labels,
        data_set=None,
        sweep_widths=None,
        spectrometer_frequencies=spectrometer_frequencies,
    )

    peak_list = PeakList(header_data, raw_peaks)

    return peak_list


def read_shift_file(gdb_file, chain_code="A"):
    data = select_records(gdb_file, VALUES)

    column_indices = get_column_indices(gdb_file)

    raw_shifts = []
    for index, line in enumerate(data, start=1):

        atom_name = line.values[column_indices["ATOMNAME"]]
        residue_number = line.values[column_indices["RESID"]]
        residue_type = line.values[column_indices["RESNAME"]]
        shift = line.values[column_indices["SHIFT"]]

        atom = AtomLabel(
            SequenceResidue(chain_code, residue_number, residue_type), atom_name
        )

        shift = ShiftData(atom, shift)

        raw_shifts.append(shift)

    return ShiftList(raw_shifts)


def _propagate_assignments(assignments):

    current = []
    result = []
    for assignment in assignments:
        fields = re.split(r"(\d+)", assignment)
        if len(fields) == 3:
            current = fields[:2]
        elif len(fields) == 1:
            fields = [*current, *fields]

        result.append(fields)

    return result


def _constructor_to_name(constructor):
    constructors_to_type_name = {int: "int", float: "float", str: "str"}

    return constructors_to_type_name[constructor]


def _constructor_names(constructors):

    result = []
    for constructor in constructors:
        result.append(_constructor_to_name(constructor))

    return result


class BadNmrPipeFile(Exception):
    """
    Base exception for bad nmr pipe files, this is the one to catch!
    """

    pass


class BadFieldFormat(BadNmrPipeFile):
    """
    One of the fields int he file has a bad format
    """

    pass


class WrongColumnCount(BadNmrPipeFile):
    """
    The number of columns in the VARS FORMAT or data lines disagree in their count
    """

    pass


class NoVarsLine(BadNmrPipeFile):
    """
    Tried to read data without a VARS line
    """

    pass


class NoFormatLine(BadNmrPipeFile):
    """
    Tried to read data without a FORMAT line
    """

    pass


class MultipleVars(BadNmrPipeFile):
    """
    Multiple VARS lines detected
    """

    pass


class MultipleFormat(BadNmrPipeFile):
    """
    Multiple FORMAT lines detected
    """

    pass


class DataBeforeFormat(BadNmrPipeFile):
    """
    Data seen before VARS and FORMAT lines detected
    """

    pass


def format_pipe_sequence(sequence_1_let: List[str]) -> str:

    """
    convert a set of 1 letter amino acid codes to an nmr pipe DATA SEQUENCE record
    :param sequence_1_let:  1 letter amino acid codes as a list of strings
    :return: nmrpipe data sequence string nicely formatted as a list of data record strings
    """

    row_strings = make_chunked_sequence_1let(sequence_1_let)

    result = []
    for row_string in row_strings:
        result.append(f"DATA SEQUENCE {row_string}")

    return result


def exit_if_required_columns_missing(
    gdb_records: DbFile, required_columns: List[str], file_name: str
):
    gdb_columns = get_gdb_columns(gdb_records)
    column_set = set(gdb_columns)

    required_column_set = set(required_columns)

    if not required_column_set.issubset(column_set):
        msg = f"""
            the required column in the file {file_name} are:

            {', '.join(required_columns)}

            found columns are

            {', '.join(gdb_columns)}

            missing columns are

            {', '.join(required_column_set - column_set)}
        """

        exit_error(msg)


def dbfile_to_first_residue_number(gdb_records: DbFile) -> int:
    first_residue_records = select_data_records(gdb_records, "FIRST_RESID")
    first_residue = 1
    if len(first_residue_records) == 1:
        first_residue = int(first_residue_records[0].values[1])
    return first_residue



from pathlib import Path
from textwrap import dedent
from typing import List

import typer

from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import exit_error, process_stream_and_add_frames
from nef_pipelines.transcoders.nmrpipe import import_app

from ...nmrview.importers.peaks import create_spectrum_frame
from ..nmrpipe_lib import (
    NMRPIPE_PEAK_EXPECTED_FIELDS,
    check_is_peak_file,
    get_gdb_columns,
    read_db_file_records,
    read_peak_file,
)

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    entry_name: str = typer.Option("nmrpipe", help="a name for the entry"),
    chain_code: str = typer.Option(
        "A", "--chain", help="chain code", metavar="<chain-code>"
    ),
    axis_codes: str = typer.Option(
        "1H.15N", metavar="<axis-codes>", help="a list of axis codes joined by dots"
    ),
    filter_noise: bool = typer.Option(True, help="remove peaks labelled as noise"),
    file_names: List[Path] = typer.Argument(
        ..., help="input peak files", metavar="<peak-file.xpk>"
    ),
):
    """convert nmrpipe peak file <nmrview>.xpk files to NEF"""

    args = get_args()

    peak_lists = _read_nmrpipe_peaks(args)

    frames = []
    for i, peak_list in enumerate(peak_lists, start=1):

        frame_name = entry_name if len(peak_lists) == 1 else f"{entry_name}_{i}"

        frames.append(create_spectrum_frame(args, frame_name, peak_list))

    entry = process_stream_and_add_frames(frames, args)

    print(entry)


def _read_nmrpipe_peaks(args):
    results = []
    for file_name in args.file_names:
        with open(file_name) as file_h:
            gdb_file = read_db_file_records(file_h, file_name=file_name)

        _check_is_peak_file_or_exit(gdb_file)

        results.append(read_peak_file(gdb_file, args))

    return results


def _check_is_peak_file_or_exit(gdb_file):
    columns = set(get_gdb_columns(gdb_file))

    if not check_is_peak_file(gdb_file):
        msg = f"""\
                this gdb file doesn't appear to contain all the columns expected for a peak file
                expected: {','.join(NMRPIPE_PEAK_EXPECTED_FIELDS)}
                got {','.join(columns & NMRPIPE_PEAK_EXPECTED_FIELDS)}
                file: {gdb_file.file_name}
                """
        msg = dedent(msg)
        exit_error(msg)



import random
import string
from collections import Counter
from dataclasses import dataclass, field
from difflib import SequenceMatcher
from enum import IntEnum, auto
from pathlib import Path
from typing import Dict, Iterable, List, Optional

from pdbx import DataContainer
from pdbx.reader import PdbxReader
from strenum import LowercaseStrEnum

from nef_pipelines.lib.structures import LineInfo
from nef_pipelines.lib.util import exit_error


class Atom:
    ...


class Chain:
    ...


class Residue:
    ...


class Model:
    ...


class Structure:
    ...


class Sequence:
    ...


@dataclass
class Atom:  # noqa: F811
    serial: int
    atom_name: str

    x: float
    y: float
    z: float

    alternative_location: str = None

    element: str = None

    temp_fact: float = None
    occupancy: float = 1.0

    residue: Optional[Residue] = None


@dataclass
class Residue:
    sequence_code: int
    residue_name: str
    atoms: List[Atom] = field(default_factory=list)
    chain: Optional[Chain] = None

    def __iter__(self):
        return self.atoms.__iter__()


@dataclass
class Chain:
    residues: List[Residue] = field(default_factory=list)
    chain_code: Optional[str] = None
    segment_id: Optional[str] = None
    sequence: Optional[Sequence] = None
    model: Optional[Model] = None

    def __iter__(self):
        return self.residues.__iter__()

@dataclass
class Model:
    serial: int
    chains: Dict[str, Chain] = field(default_factory=dict)
    structure: Optional[Structure] = None

    def __iter__(self):
        return iter(self.chains.values())


class PdbSecondaryStructureType(IntEnum):
    HELIX_RIGHT_HANDED_ALPHA = auto()
    HELIX_RIGHT_HANDED_OMEGA = auto()
    HELIX_RIGHT_HANDED_PI = auto()
    HELIX_RIGHT_HANDED_GAMMA = auto()
    HELIX_RIGHT_HANDED_3_10 = auto()
    HELIX_LEFT_HANDED_ALPHA = auto()
    HELIX_LEFT_HANDED_OMEGA = auto()
    HELIX_LEFT_HANDED_GAMMA = auto()
    HELIX_RIBBON_HELIX_2_7 = auto()
    HELIX_POLYPROLINE = auto()
    SHEET = auto()


@dataclass
class SecondaryStructureElement:
    chain_code: str
    start_sequence_code: int
    end_sequence_code: int
    alternative_location: str
    secondary_structure_type: PdbSecondaryStructureType

    structure: Optional[Structure] = None


@dataclass
class Sequence:
    id: int
    start_sequence_code: int
    residues: List[str] = field(default_factory=list)

    structure: Optional[Structure] = None


@dataclass
class Structure:
    source: str = None
    sequences: Dict[str, Sequence] = field(default_factory=dict)
    secondary_structure: Dict[str, List[SecondaryStructureElement]] = field(
        default_factory=dict
    )

    models: List[Model] = field(default_factory=list)

    def __iter__(self):
        return self.models.__iter__()


    def __getitem__(self, index):
        return self.models[0]


current_structure: Optional[Structure] = None
current_model: Optional[Model] = None
current_chain: Optional[Chain] = None
current_residue: Optional[Residue] = None
current_indices: Dict[str, int] = {}
current_lines: List[str]


def _pad_line_to_80(line):
    line_length = len(line)
    if line_length < 80:
        line += " " * (80 - line_length)
    return line


@dataclass
class PDBLineInfo(LineInfo):
    record_type: str


def _convert_to_int_or_exit(
    value_string: str, line_info: PDBLineInfo, field_name: str
) -> int:
    converter = int
    return _convert_to_type_or_exit_error(
        value_string, converter, line_info, field_name
    )


def _convert_to_float_or_exit(
    value_string: str, line_info: PDBLineInfo, field_name: str
) -> int:
    converter = float
    return _convert_to_type_or_exit_error(
        value_string, converter, line_info, field_name
    )


def _convert_to_type_or_exit_error(value_string, converter, line_info, field_name):
    result = None
    try:
        result = converter(value_string)
    except ValueError:
        conversion_type = str(converter).split()[1].strip(">").strip("'")
        _report_bad_conversion(
            value_string,
            conversion_type,
            line_info,
            field_name,
        )
    return result


def _report_bad_conversion(
    value: str, conversion_type: str, line_info: PDBLineInfo, field_name: str
):
    msg = f"""
            at line {line_info.line_no} in {line_info.file_name}
            the field {field_name} in a {line_info.record_type} record
            couldn't be converted to a {conversion_type}

            the value of the field {field_name} was |{value}|

                        ruler 10s |         1         2         3         4         5         6         7         8|
                        ruler 1s  |12345678901234567890123456789012345678901234567890123456789012345678901234567890|
            the complete line was |{line_info.line}|

            note all |'s are used to delimit the start and end of lines and values and are not part of the value
        """
    exit_error(msg)


def _as_string_or_none(item):
    if item:
        item = item.strip()
        if not item:
            item = None
    return item


def _as_continuous_string_or_exit(name, line_info: PDBLineInfo, field_name: str):
    name = name.strip()

    if string.whitespace in name:
        _report_bad_conversion(name, "string without spaces", line_info, field_name)

    return name


def _exit_if_no_chain_code_and_no_segment_id(chain_code, segment_id, line_info):
    if not chain_code and not segment_id:
        msg = f"""
            in file {line_info.file_name} at line {line_info.line_no} both the chain code and segment id were
            not present on an ATOM record, one or both must be present

            the line was

            {line_info.line}
        """

        exit_error(msg)


def _parse_atom(line, line_info):
    global current_residue, current_chain, current_model

    serial = line[6:11]
    name = line[12:16]
    alternative_location = line[16]
    residue_name = line[17:20]
    chain_code = line[21]
    sequence_code = line[22:26]
    x = line[30:38]
    y = line[38:46]
    z = line[46:54]
    segment_id = line[72:76]
    element = line[76:78]
    temp_fact = line[60:66]

    serial = _convert_to_int_or_exit(serial, line_info, "serial")
    sequence_code = _convert_to_int_or_exit(sequence_code, line_info, "sequence_code")

    x = _convert_to_float_or_exit(x, line_info, "x")
    y = _convert_to_float_or_exit(y, line_info, "y")
    z = _convert_to_float_or_exit(z, line_info, "z")

    segment_id = _as_string_or_none(segment_id)
    chain_code = _as_string_or_none(chain_code)

    element = _as_string_or_none(element)

    temp_fact = _convert_to_float_or_exit(temp_fact, line_info, "temperature factor")

    name = _as_continuous_string_or_exit(name, line_info, "name")

    alternative_location = _as_string_or_none(alternative_location)

    residue_name = _as_continuous_string_or_exit(
        residue_name, line_info, "residue name"
    )

    current_atom = Atom(
        serial=serial,
        atom_name=name,
        alternative_location=alternative_location,
        x=x,
        y=y,
        z=z,
        element=element,
        temp_fact=temp_fact,
    )

    _exit_if_chain_code_and_segid_are_mismatched(
        current_chain, chain_code, segment_id, line_info
    )

    _exit_if_no_chain_code_and_no_segment_id(chain_code, segment_id, line_info)

    if current_chain:
        new_chain = False
        if (current_chain.chain_code and chain_code) and current_chain.chain_code != chain_code:
            new_chain = True

        if current_chain.segment_id and segment_id and current_chain.segment_id != segment_id:
            new_chain = True

        if new_chain:
            current_chain = None

    if not current_chain:
        chain_segment_key = chain_code if chain_code else segment_id
        current_chain = Chain(chain_code=chain_code, segment_id=segment_id)
        current_model.chains[chain_segment_key] = current_chain

    if current_residue and current_residue.sequence_code != sequence_code:
        current_residue = None

    if not current_residue:
        current_residue = Residue(
            sequence_code=sequence_code, residue_name=residue_name, chain=current_chain
        )
        current_chain.residues.append(current_residue)

    current_atom.residue = current_residue
    current_residue.atoms.append(current_atom)


def _exit_if_chain_code_and_segid_are_mismatched(
    target_chain, chain_code, segment_id, line_info
):
    mismatch = None
    non_mismatch = None

    if (
        target_chain
        and target_chain.chain_code == chain_code
        and chain_code
        and target_chain.segment_id != segment_id
    ):
        mismatch = "segment id"
        non_mismatch = " chain code"

    if (
        target_chain
        and target_chain.segment_id == segment_id
        and segment_id
        and target_chain.chain_code != chain_code
    ):
        mismatch = "chain code"
        non_mismatch = "segment id"

    if mismatch:
        msg = f"""
            while reading a chain from the file {line_info.file_name} at line {line_info.line_no}
            the {mismatch} changed but the {non_mismatch} didn't

            chain code previous: {target_chain.chain_code} current: |{chain_code}|
            segment id previous: {target_chain.segment_id} current: |{segment_id}|

            the current line is

            {line_info.line}

            note |'s are to delimit whitespace in the segment id and are not part of the segment id itself
                 a segment id is always 4 characters long
        """

        exit_error(msg)


def _parse_sequence(line: str, _: PDBLineInfo):
    global current_structure
    chain_code = line[11]
    chain_code = _as_string_or_none(chain_code)

    sequence = current_structure.sequences.setdefault(chain_code, Sequence(None, None))

    for offset in range(13):
        start = 19 + (offset * 4)
        end = 22 + (offset * 4)

        target_residue = line[start:end].strip()
        if not target_residue:
            break
        sequence.residues.append(target_residue)


def _parse_helix(line: str, line_info: PDBLineInfo):
    global current_structure

    chain_code = line[19]
    alternative_location = line[25]
    first_sequence_code = line[21:25]
    last_sequence_code = line[33:37]
    helix_type = line[38:40]

    chain_code = chain_code.strip()
    alternative_location = _as_string_or_none(alternative_location)
    first_sequence_code = _convert_to_int_or_exit(
        first_sequence_code, line_info, "helix.initial-sequence-number"
    )
    last_sequence_code = _convert_to_int_or_exit(
        last_sequence_code, line_info, "helix.last-sequence-number"
    )
    helix_type_offset = _convert_to_int_or_exit(helix_type, line_info, "helix.class")
    secondary_structure_type = PdbSecondaryStructureType(helix_type_offset)

    secondary_structure_element = SecondaryStructureElement(
        chain_code,
        first_sequence_code,
        last_sequence_code,
        alternative_location,
        secondary_structure_type,
    )

    current_structure.secondary_structure.setdefault(chain_code, []).append(
        secondary_structure_element
    )


def _parse_sheet(line, line_info):
    global current_structure

    chain_code = line[21]
    first_sequence_code = line[22:26]
    alternative_location = line[26]
    last_sequence_code = line[33:37]

    first_sequence_code = _convert_to_int_or_exit(
        first_sequence_code, line_info, "start sequence code"
    )
    last_sequence_code = _convert_to_int_or_exit(
        last_sequence_code, line_info, "end sequence_code"
    )
    alternative_location = _as_string_or_none(alternative_location)

    secondary_structure_element = SecondaryStructureElement(
        chain_code,
        first_sequence_code,
        last_sequence_code,
        alternative_location,
        PdbSecondaryStructureType.SHEET,
    )

    current_structure.secondary_structure.append(secondary_structure_element)


def _fixup_sequences(current_structure):
    sequence_index_by_sequences = {}
    sequence_count = 1
    for sequence in current_structure.sequences.values():
        sequence = tuple(sequence.residues)
        if sequence not in sequence_index_by_sequences:
            sequence_index_by_sequences[sequence] = str(sequence_count)
            sequence_count += 1

    sequences = {
        index: Sequence(index, None, list(sequence), current_structure)
        for sequence, index in sequence_index_by_sequences.items()
    }

    current_structure.sequences = sequences


def _fixup_secondary_structure(current_structure):
    for secondary_structure_list in current_structure.secondary_structure.values():
        for secondary_structure in secondary_structure_list:
            secondary_structure.structure = current_structure


def parse_pdb(lines: Iterable[str], source: str = "unknown"):
    global current_model, current_chain, current_residue, current_structure

    current_structure = Structure(source)

    for line_no, line in enumerate(lines, start=1):

        line = line.rstrip("\n")

        line = _pad_line_to_80(line)

        record_type = line.split()[0]

        line_info = PDBLineInfo(
            source, line_no=line_no, line=line, record_type=record_type
        )

        if record_type == "ATOM":
            if not current_model:
                current_model = Model(1)
                current_structure.models.append(current_model)
            _parse_atom(line, line_info)

        if record_type == "TER":
            if not current_chain:
                msg = f"""
                    at line {line_info.line_no} in {line_info.file_name}
                    there was a termination line when no chain was bein read

                    the line was

                    {line_info.line}
                """

                exit_error(msg)

            current_chain = None
            current_residue = None

        if record_type == "MODEL":
            model_number = line[10:14]
            model_number = _convert_to_int_or_exit(model_number, line_info, "MODEL")

            if current_model:
                msg = f"""
                    at line {line_info.line_no} in the file {line_info.file_name}
                    a new model started when a model was already open
                    the new model number was {model_number} the old model_numbers was {current_model.serial}

                    the line was

                    |{line_info.line}|

                    note |'s are to delimit whitespace in the line and are not part of the line itself
                """

                exit_error(msg)

            current_model = Model(model_number)
            current_structure.models.append(current_model)

        if record_type == "ENDMDL":
            current_model = None
            current_chain = None
            current_residue = None

        if record_type == "SEQRES":
            _parse_sequence(line, line_info)

        if record_type == "HELIX":
            _parse_helix(line, line_info)

        if record_type == "SHEET":
            _parse_sheet(line, line_info)

    _fixup_sequences(current_structure)
    _fixup_secondary_structure(current_structure)

    _match_sequences_and_set_offsets(current_structure)

    for secondary_structure_list in current_structure.secondary_structure.values():
        secondary_structure_list.sort(key=lambda x: x.start_sequence_code)

    result = current_structure
    current_structure = None
    current_model = None
    current_chain = None
    current_residue = None

    return result


def _find_last_row_indices(lines, targets):
    result = {}
    for i, line in enumerate(lines):
        for target in targets:
            if target in line:
                result[target] = i

    return result


def _find_row(line_data, target):
    global current_lines, current_indices

    offset = current_indices[target]

    print(offset)


def _attibute_index_to_name(items, target_index):
    result = None
    for name, index in items.attribute_list_with_order:
        if index == target_index:
            result = name
            break
    return result


class ComputedLineInfo:
    def __init__(self, lines: List[str], file_name: str, search_window=10):
        self._lines: List[str] = lines
        self.file_name = file_name
        self._container_offset = None
        self._line_number = None
        self._search_window = search_window
        self._search_terms = None

    def set_container(self, container: str):
        for line_number, line in enumerate(self._lines, start=1):

            if line.strip().startswith(f"_{container}."):
                self._container_offset = line_number
                self._line_number = f"close to / below {self._container_offset}"

    def set_container_row(self, offset):
        self._line_number = self._container_offset + offset + 1
        self._search_terms = None

    @property
    def line(self):
        return (
            self._lines[self._line_number]
            if self._line_number
            else "value of line unknown"
        )

    @property
    def line_no(self):

        result = self._line_number

        test_line = self._line_number if self._line_number else None

        if self._search_terms:

            start = test_line - self._search_window
            end = test_line + self._search_window
            scores = self.score_lines(end, start)

            failed = self._scores_good(scores)

            if failed:
                start = self._container_offset
                end = len(self._lines) - 1
                scores = self._score_lines(start, end)

            failed = self._scores_good(scores)

            if not failed:
                result = scores.most_common()[0][0]

        return result

    def _scores_good(self, scores):
        failed = False
        if len(scores) > 1:
            failed = False
            if scores.most_common()[0][1] == scores.most_common()[1][1]:
                failed = True

            if scores.most_common()[0][1] != len(self._search_terms):
                failed = True
        return failed

    def score_lines(self, end, start):
        scores = Counter()
        for i, search_line_line_number in enumerate(range(start, end)):

            if search_line_line_number > 0 and search_line_line_number < (
                len(self._lines) - 1
            ):
                fields = self._lines[search_line_line_number].split()
                for search_term in self._search_terms:
                    print(
                        i,
                        search_line_line_number,
                        str(search_term) in fields,
                        search_term,
                        fields,
                    )
                    if str(search_term) in fields:
                        scores[search_line_line_number + 1] += 1
        return scores

    def set_search_terms(self, *args):
        self._search_terms = list(args)


def _parse_cif_atoms(data: DataContainer, line_info: ComputedLineInfo) -> Structure:
    global current_structure, current_model, current_chain, current_residue

    atoms = data.get_object("atom_site")
    line_info.set_container("atom_site")

    record_type_index = atoms.get_attribute_index("group_PDB")
    atom_id_index = atoms.get_attribute_index("id")
    element_index = atoms.get_attribute_index("type_symbol")

    seq_code_index = _get_attribute_index_favour_auth(atoms, "{source}_seq_id")
    residue_name_index = _get_attribute_index_favour_auth(atoms, "{source}_comp_id")
    chain_code_index = _get_attribute_index_favour_auth(atoms, "{source}_asym_id")
    atom_name_index = _get_attribute_index_favour_auth(atoms, "{source}_atom_id")

    alternative_location_index = atoms.get_attribute_index("pdbx_PDB_ins_code")

    model_index = atoms.get_attribute_index("pdbx_PDB_model_num")
    occupancy_index = atoms.get_attribute_index("occupancy")
    x_index = atoms.get_attribute_index("Cartn_x")
    y_index = atoms.get_attribute_index("Cartn_y")
    z_index = atoms.get_attribute_index("Cartn_z")
    B_iso_or_equiv_index = atoms.get_attribute_index("B_iso_or_equiv")

    current_model = None
    for i, row in enumerate(atoms, start=1):
        line_info.set_container_row(i)
        line_info.set_search_terms(*row)

        record_type = row[record_type_index]

        if record_type == "ATOM":
            atom_id = row[atom_id_index]
            element = row[element_index]
            chain_code = row[chain_code_index]
            atom_name = row[atom_name_index]
            sequence_code = row[seq_code_index]
            alternative_location = row[alternative_location_index]
            residue_name = row[residue_name_index]
            model_number = row[model_index]
            occupancy = row[occupancy_index]
            x = row[x_index]
            y = row[y_index]
            z = row[z_index]
            b_iso_or_equiv = row[B_iso_or_equiv_index]

            seq_code_attribute_name = _attibute_index_to_name(atoms, seq_code_index)

            # use correct line info and correct name in conversion of sequence_id
            atom_id = _convert_to_int_or_exit(atom_id, None, "atom_id")
            element = _as_string_or_none(element)
            sequence_code = _convert_to_int_or_exit(
                sequence_code,
                line_info,
                f"sequence atom_id [{seq_code_attribute_name}]",
            )
            alternative_location = _as_string_or_none(alternative_location)
            model_number = _convert_to_int_or_exit(model_number, line_info, "model")
            occupancy = _convert_to_float_or_exit(occupancy, line_info, "occupancy")
            x = _convert_to_float_or_exit(x, line_info, "x")
            y = _convert_to_float_or_exit(y, line_info, "y")
            z = _convert_to_float_or_exit(z, line_info, "z")

            b_iso_or_equiv = _as_string_or_none(b_iso_or_equiv)
            if b_iso_or_equiv:
                b_iso_or_equiv = _convert_to_float_or_exit(
                    b_iso_or_equiv, line_info, "b value"
                )

            if not current_model or model_number != current_model.serial:
                current_model = Model(model_number, structure=current_structure)
                current_structure.models.append(current_model)
                current_chain = None

            if current_chain and current_chain.chain_code != chain_code:
                current_chain = None
                current_residue = None

            sequence = (
                current_structure.sequences[sequence_code]
                if sequence_code in current_structure.sequences
                else None
            )

            if not current_chain:
                current_chain = Chain(
                    chain_code=chain_code, sequence=sequence, model=current_model
                )
                current_model.chains[chain_code] = current_chain

            if current_residue and current_residue.sequence_code != sequence_code:
                current_residue = None

            if not current_residue:
                current_residue = Residue(
                    sequence_code, residue_name, chain=current_chain
                )
                current_chain.residues.append(current_residue)

            current_atom = Atom(
                atom_id,
                atom_name,
                x,
                y,
                z,
                alternative_location,
                element,
                b_iso_or_equiv,
                occupancy,
                current_residue,
            )

            current_residue.atoms.append(current_atom)

    current_model = None
    current_chain = None
    current_residue = None

    result = current_structure
    current_structure = None

    return result


def _get_attribute_index_favour_auth(atoms, attribute_template):
    auth_attribute_name = attribute_template.format(source="auth")
    label_attribute_name = attribute_template.format(source="label")
    if atoms.has_attribute(auth_attribute_name):
        result = atoms.get_attribute_index(auth_attribute_name)
    else:
        result = atoms.get_attribute_index(label_attribute_name)

    return result


def _parse_cif_helix(data, line_info):
    global current_structure

    if helices := data.get_object("struct_conf"):
        line_info.set_container("struct_conf")

        chain_code_index = _get_attribute_index_favour_auth(
            helices, "beg_{source}_asym_id"
        )
        secondary_start_index = _get_attribute_index_favour_auth(
            helices, "beg_{source}_seq_id"
        )
        secondary_end_index = _get_attribute_index_favour_auth(
            helices, "end_{source}_seq_id"
        )
        alternative_location_index = helices.get_attribute_index(
            "pdbx_beg_PDB_ins_code"
        )
        helix_type_index = helices.get_attribute_index("pdbx_PDB_helix_class")

        for i, line in enumerate(helices):
            line_info.set_container_row(i)
            line_info.set_search_terms(*line)

            chain_code = line[chain_code_index]
            alternative_location = line[alternative_location_index]
            first_sequence_code = line[secondary_start_index]
            last_sequence_code = line[secondary_end_index]
            helix_type = line[helix_type_index]

            alternative_location = _as_string_or_none(alternative_location)
            first_sequence_code = _convert_to_int_or_exit(
                first_sequence_code, line_info, "helix.initial-sequence-number"
            )
            last_sequence_code = _convert_to_int_or_exit(
                last_sequence_code, line_info, "helix.last-sequence-number"
            )
            helix_type_offset = _convert_to_int_or_exit(
                helix_type, line_info, "helix.class"
            )
            secondary_structure_type = PdbSecondaryStructureType(helix_type_offset)

            secondary_structure_element = SecondaryStructureElement(
                chain_code,
                first_sequence_code,
                last_sequence_code,
                alternative_location,
                secondary_structure_type,
                structure=current_structure,
            )

            current_structure.secondary_structure.setdefault(chain_code, []).append(
                secondary_structure_element
            )


def _parse_cif_sheet(data, line_info):
    global current_structure

    if sheets := data.get_object("struct_sheet_range"):
        line_info.set_container("struct_sheet_range")

        chain_code_index = _get_attribute_index_favour_auth(
            sheets, "beg_{source}_asym_id"
        )
        secondary_start_index = _get_attribute_index_favour_auth(
            sheets, "beg_{source}_seq_id"
        )
        secondary_end_index = _get_attribute_index_favour_auth(
            sheets, "end_{source}_seq_id"
        )
        alternative_location_index = sheets.get_attribute_index("pdbx_beg_PDB_ins_code")

        for i, line in enumerate(sheets):
            line_info.set_container_row(i)
            line_info.set_search_terms(line)
            chain_code = line[chain_code_index]
            first_sequence_code = line[secondary_start_index]
            last_sequence_code = line[secondary_end_index]
            alternative_location = line[alternative_location_index]

            first_sequence_code = _convert_to_int_or_exit(
                first_sequence_code, None, "start sequence code"
            )
            last_sequence_code = _convert_to_int_or_exit(
                last_sequence_code, None, "end sequence_code"
            )
            alternative_location = _as_string_or_none(alternative_location)

            secondary_structure_element = SecondaryStructureElement(
                chain_code,
                first_sequence_code,
                last_sequence_code,
                alternative_location,
                PdbSecondaryStructureType.SHEET,
            )

            current_structure.secondary_structure.setdefault(chain_code, []).append(
                secondary_structure_element
            )


def parse_cif(fh: Iterable[str], source: str) -> Structure:
    global current_structure, current_indices

    current_lines = fh.readlines()

    current_indices = _find_last_row_indices(
        current_lines,
        [
            "_atom_site",
        ],
    )

    current_structure = Structure(source)

    reader = PdbxReader(current_lines)
    data = []
    reader.read(data)
    data = data[0]

    line_info = ComputedLineInfo(current_lines, file_name=source)

    _parse_cif_sequence(data, line_info)
    _parse_cif_helix(data, line_info)
    _parse_cif_sheet(data, line_info)
    structure = _parse_cif_atoms(data, line_info)

    _match_sequences_and_set_offsets(structure)

    for secondary_structure_list in structure.secondary_structure.values():
        secondary_structure_list.sort(key=lambda x: x.start_sequence_code)

    return structure


def _match_sequences_and_set_offsets(structure):
    if structure.sequences:
        matcher = SequenceMatcher()
        chain_matches = {}
        chain_starts = {}
        for chain in structure.models[0].chains.values():
            chain_segment_id_key = chain.chain_code if chain.chain_code else chain.segment_id
            chain_residues = {}
            chain_starts[chain_segment_id_key] = chain.residues[0].sequence_code

            for residue in chain.residues:
                chain_residues[residue.sequence_code] = residue.residue_name


            min_residue = min(chain_residues.keys())
            max_residue = max(chain_residues.keys())

            match_residues = []
            for i in range(min_residue, max_residue + 1):
                match_residue = chain_residues[i] if i in chain_residues else "."
                match_residues.append(match_residue)

            for index, sequence in structure.sequences.items():
                matcher.set_seqs(match_residues, sequence.residues)
                chain_segment_key = chain.chain_code if chain.chain_code else  chain.segment_id
                chain_matches.setdefault(chain_segment_key, {})[matcher.ratio()] = (
                    index,
                    matcher.get_opcodes(),
                )

        chain_offsets = {}
        for chain, matches in chain_matches.items():
            best = max(matches)
            best_matches = matches[best]
            for item in best_matches[1]:
                if item[0] == "equal":

                    chain_offsets[chain] = best_matches[0], item[3]
                    break

        seen_sequences = set()

        for chain, (sequence_id, offset) in chain_offsets.items():
            if sequence_id not in seen_sequences:
                seen_sequences.add(sequence_id)
                chain_first_residue = (
                    structure.models[0].chains[chain].residues[0].sequence_code
                )
                structure.sequences[sequence_id].start_sequence_code = (
                    int(chain_first_residue) - offset
                )


def _parse_cif_sequence(data, line_info):
    global current_structure, current_indices

    sequence = data.get_object("entity_poly_seq")
    line_info.set_container("entity_poly_seq")

    entity_id_index = sequence.get_attribute_index("entity_id")
    monomer_id_index = sequence.get_attribute_index("mon_id")

    for i, row in enumerate(sequence.row_list, start=1):
        line_info.set_container_row(i)
        line_info.set_search_terms(*row)

        entity_id = row[entity_id_index]
        monomer_id = row[monomer_id_index]
        current_structure.sequences.setdefault(
            entity_id, Sequence(entity_id, None, structure=current_structure)
        )

        current_structure.sequences[entity_id].residues.append(monomer_id)

PDB_RECORD_IDS = set([

    'HEADER', 'OBSLTE', 'TITLE', 'SPLT',
    'CAVEAT', 'COMPND', 'SOURCE','KEYWDS',
    'EXPDTA','NUMMDL','MDLTYP','AUTHOR',
    'REVDAT','SPRSDE','JRNL', 'REMARKS',
    'DBREF', 'DBREF1', 'DBREF2', 'SEQADV',
    'SEQRES', 'MODRES', 'HET','FORMUL',
    'HETNAM','HETSYN', 'HELIX', 'SHEET',
    'SSBOND', 'LINK', 'CISPEP', 'SITE'
    'CRYST1', 'MTRIXn', 'ORIGXn', 'SCALEn',
    'MODEL', 'ANISOU', 'TER', 'ENDMDL',
    'CONECT', 'MASTER', 'END'
])

class RCSBFileType(LowercaseStrEnum):
    PDB= auto(),
    CIF = auto(),
    UNKNOWN =  auto()

def guess_cif_or_pdb(lines: Iterable[str], file_name: str = '', test_length: int =100):

    pdb = 0
    cif = 0

    file_path = Path(file_name)

    if file_path.suffix.lower() in ('.cif', '.mmcif', '.pdbx'):
        cif = 1
    elif file_path.suffix.lower() == '.pdb':
        pdb = 1
    else:
        for line in lines[:test_length]:
            fields = line.strip().split()
            if fields[0] in PDB_RECORD_IDS:
                pdb+=1
            if fields[0].startswith('data_') or fields[0][0] in ('_', ';') or fields[0] == 'loop_':
                cif+=1

    if pdb > cif:
        result = RCSBFileType.PDB
    elif cif > pdb:
        result = RCSBFileType.CIF
    else:
        result = RCSBFileType.UNKNOWN

    return result

if __name__ == "__main__":
    root = Path(
        "/Users/garythompson/Dropbox/nef_pipelines/nef_pipelines/src/nef_pipelines/tests/rcsb/test_data/"
    )
    file_name = random.choice(
        [
            root / Path("1l2y_short.cif"),
            root / Path("1l2y_short.pdb"),
        ]
    )
    file_path = Path(file_name)
    fh = open(file_path)

    file_name = file_path.parts[-1]
    if file_path.suffix == ".cif":
        structure = parse_cif(fh, file_name)
    else:
        structure = parse_pdb(fh, file_name)

    print(structure.sequences)
    print(structure.secondary_structure)
    for model in structure.models:
        print(len(model.chains))
        for chain in model.chains:
            for residue in chain.residues:
                for atom in residue:
                    print(
                        model.serial,
                        "|",
                        chain.chain_code,
                        chain.segment_id,
                        "|",
                        residue.sequence_code,
                        residue.residue_name,
                        "|",
                        atom.serial,
                        atom.atom_name,
                        atom.alternative_location,
                        atom.x,
                        atom.y,
                        atom.z,
                        atom.element,
                        atom.temp_fact,
                        atom.occupancy,
                    )
    print(structure.secondary_structure)
    print(structure.sequences)



import typer

import nef_pipelines
from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:

    nef_app.app.add_typer(app, name="pales", help="-  read and write pales/dc [rdcs]")

    # app.add_typer(import_app, name='import', help='-  import pales/dc [rdc restraints]')
    app.add_typer(export_app, name="export", help="- export pales/dc [rdc fits]")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.pales.exporters.rdcs  # noqa: F401
    import nef_pipelines.transcoders.pales.exporters.template  # noqa: F401



import sys
from fnmatch import fnmatch
from pathlib import Path
from textwrap import dedent, indent
from typing import List

import typer
from fyeah import f
from Levenshtein import distance
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames_by_name,
)
from nef_pipelines.lib.util import (
    FOUR_SPACES,
    STDIN,
    chunks,
    exit_error,
    parse_comma_separated_options,
    strings_to_table_terminal_sensitive,
)
from nef_pipelines.tools.frames import frames_app

# TODO: add mmv like semantics as an option [https://manpages.ubuntu.com/manpages/bionic/man1/mmv.1.html]

REPLACE_HELP = """replace part of the frame name selected with the replacement string, rather than the whole name"""


@frames_app.command()
def rename(
    input: Path = typer.Option(
        STDIN, "-i", "--in", help="file to read input from [- is stdin]"
    ),
    exact: bool = typer.Option(False),
    replace: bool = typer.Option(False, help=REPLACE_HELP),
    delete: bool = typer.Option(
        False, help="delete the text rather than replacing it if replace is specified"
    ),
    category: str = typer.Option(
        "*", help="category to match as well as the frame name"
    ),
    rename_category: bool = typer.Option(False, help="change the category"),
    force: bool = typer.Option(
        False,
        help="force a replacement so a rename can delete a frame with a name clash",
    ),
    old_new_names: List[str] = typer.Argument(
        ...,
        help="a list of pairs of old and new names for frames, arguments can be repeated or comma separated",
    ),
):
    """- rename frames in the current input"""

    if delete and not replace:
        exit_error("you can't use --delete without --replace")

    old_new_names = parse_comma_separated_options(old_new_names)

    if delete:
        new_new_old_names = []
        for name in old_new_names:
            new_new_old_names.append(name)
            new_new_old_names.append("")
        old_new_names = new_new_old_names

    _exit_renames_not_pairs(old_new_names)

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    for old_name, new_name in chunks(old_new_names, 2):

        target_frames = select_frames_by_name(
            entry,
            [
                old_name,
            ],
            exact,
        )

        if category != "*":
            if exact:
                target_frames = [
                    frame for frame in target_frames if frame.category == category
                ]
            else:
                target_frames = [
                    frame
                    for frame in target_frames
                    if fnmatch(frame.category, f"*{category}*")
                ]

        if len(target_frames) == 0:
            _exit_no_frames_selected(old_name, category, entry, exact)

        if len(target_frames) > 1 and not replace:
            _exit_error_mutiple_frames_selected(
                old_name, category, exact, entry, target_frames
            )

        target_frame = target_frames[0]

        # TODO: check for accceptable characters [33-126] though maybe we allow unicode
        # TODO: tests
        for target_frame in target_frames:
            if len(new_name.split()) > 1:
                _exit_spaces_in_new_name(new_name)

            if not rename_category:
                category = target_frame.category
                new_name_part = new_name
                if replace:
                    frame_name = target_frame.name[len(category) :].lstrip("_")
                    new_name_part = frame_name.replace(old_name, new_name)
                    if new_name_part.startswith("_"):
                        new_name_part = new_name_part[1:]

                new_full_name = f"{category}_{new_name_part}"

                if new_full_name in entry.frame_dict.keys() and not force:
                    _exit_clashing_frame_name(new_full_name, entry)

                if new_full_name in entry.frame_dict.keys() and force:
                    frame_to_remove = entry.get_saveframe_by_name(new_full_name)
                    entry.remove_saveframe(frame_to_remove)

                target_frame.name = new_full_name

            else:

                name_part = target_frame.name[len(target_frame.category) :].lstrip("_")
                new_full_name = f"{new_name}_{name_part}"

                target_frame.category = new_name
                target_frame.name = new_full_name

    print(entry)


def _exit_error_mutiple_frames_selected(
    target_name, target_category, exact, entry, target_frames
):
    if not exact:
        target_name = f"*{target_name}*"
        target_category = f"*{target_category}*"

    category_msg = (  # noqa: F841
        "" if target_category == "*" else f" and the category {target_category}"
    )

    target_frame_names = [frame.name for frame in target_frames]
    frames_table = strings_to_table_terminal_sensitive(target_frame_names)
    frames_table = tabulate(frames_table, tablefmt="plain")
    frames = indent(dedent(frames_table), FOUR_SPACES)  # noqa: F841

    msg = """
        multiple save frames were selected using the name {target_name}{category_msg} in entry {entry.entry_id}
        and I can only rename one save frame at the same time...

        the save frames were

        {frames}
    """

    msg = dedent(msg)
    exit_error(f(msg))


def _exit_renames_not_pairs(old_new_names):
    if len(old_new_names) % 2 != 0:
        pairs = ", ".join(old_new_names[:-1])
        pairs = indent(pairs, "   ")
        msg = f"""\
            old and new names must be pairs, i got an extra term {old_new_names[-1]}'

            pairs were:

            {pairs}
        """

        exit(msg)

        print(msg, file=sys.stderr)
        print(file=sys.stderr)


def _exit_no_frames_selected(target_name, target_category, entry, exact):
    if not exact:
        target_name = f"*{target_name}*"
        target_category = f"*{target_category}*"

    all_names_and_categories = [(frame.name, frame.category) for frame in entry]

    if len(all_names_and_categories) == 0:
        msg = """
            there were no frames in the entry to rename
        """
    else:
        category_msg = (
            "" if target_category == "*" else f" with category {target_category}"
        )
        category_msg = category_msg
        distances = [
            (
                distance(name, target_name) + distance(target_category, category),
                (name, category),
            )
            for name, category in all_names_and_categories
        ]
        distances.sort()

        all_names = [name for name, _ in all_names_and_categories]
        table = strings_to_table_terminal_sensitive(all_names)
        table = tabulate(table, tablefmt="plain")
        table = indent(dedent(table), FOUR_SPACES)

        template = """
            the frame {target_name}{category_msg} wasn't found in the entry {entry.entry_id},
            did you mean {distances[0][-1][0]} [category: {distances[0][-1][1]}]?

            all the frame names in the entry {entry.entry_id} were:

            {table}
        """
        template = dedent(template)

        msg = f(template)

    exit_error(msg)


def _exit_spaces_in_new_name(new_name):
    msg = f"""
        frame names can't contain spaces your new name was  {new_name} and contains spaces
    """

    exit_error(msg)


def _exit_clashing_frame_name(new_full_name, entry):
    msg = f"""
        a frame with the name {new_full_name} already exists in entry {entry.entry_id}  you will need to use the --force
        option to force the rename to replace it (if that's what you want to do!)
    """

    exit_error(msg)



import typer

from nef_pipelines import nef_app

chains_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        chains_app, name="chains", help="-  carry out operations on chains"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.chains.clone  # noqa: F401
    import nef_pipelines.tools.chains.list  # noqa: F401
    import nef_pipelines.tools.chains.rename  # noqa: F401
    import nef_pipelines.tools.chains.renumber  # noqa: F401
